@@ -1,6 +1,6 @@
 #include "utils/profiler.hpp"
 #include <sirius.hpp>
-#include "filesystem.hpp"
+#include "utils/filesystem.hpp"
 #include <utils/json.hpp>
 #include <cfenv>
 #include <fenv.h>
@@ -15,7 +15,8 @@ enum class task_t : int
     ground_state_new     = 0,
     ground_state_restart = 1,
     k_point_path         = 2,
-    eos                  = 3
+    eos                  = 3,
+    read_config          = 4
 };
 
 void json_output_common(json& dict__)
@@ -302,11 +303,11 @@ void run_tasks(cmd_args const& args)
     if (task == task_t::ground_state_new || task == task_t::ground_state_restart) {
         auto ctx = create_sim_ctx(fname, args);
         ctx->initialize();
-        if (ctx->comm().rank() == 0) {
-            auto dict = ctx->serialize();
-            std::ofstream ofs("setup.json", std::ofstream::out | std::ofstream::trunc);
-            ofs << dict.dump(4);
-        }
+        //if (ctx->comm().rank() == 0) {
+        //    auto dict = ctx->serialize();
+        //    std::ofstream ofs("setup.json", std::ofstream::out | std::ofstream::trunc);
+        //    ofs << dict.dump(4);
+        //}
         //if (ctx->full_potential()) {
         //    ctx->gk_cutoff(ctx->aw_cutoff() / ctx->unit_cell().min_mt_radius());
         //}
@@ -339,6 +340,35 @@ void run_tasks(cmd_args const& args)
             }
         }
     }
+    if (task == task_t::read_config) {
+        //int count{0};
+        //while (true) {
+        //    std::stringstream s;
+        //    s << "sirius" << std::setfill('0') << std::setw(6) << count << ".json";
+        //    fname = s.str();
+        //    try {
+        //        auto dict = utils::read_json_from_file_or_string(fname);
+        //        Simulation_context ctx(dict["config"].dump());
+        //    } catch(...) {
+        //        break;
+        //    }
+        //    count++;
+        //}
+        auto dict1 = utils::read_json_from_file_or_string("sirius000030.json");
+        Simulation_context ctx1(dict1["config"].dump());
+        ctx1.initialize();
+
+        auto dict2 = utils::read_json_from_file_or_string("sirius000031.json");
+        Simulation_context ctx2(dict1["config"].dump());
+        ctx2.initialize();
+
+        ctx1.unit_cell().set_lattice_vectors(ctx2.unit_cell().lattice_vectors());
+        for (int ia = 0; ia < ctx2.unit_cell().num_atoms(); ia++) {
+            ctx1.unit_cell().atom(ia).set_position(ctx2.unit_cell().atom(ia).position());
+        }
+        ctx1.update();
+
+    }
 
     if (task == task_t::k_point_path) {
         auto ctx = create_sim_ctx(fname, args);
@@ -397,7 +427,7 @@ void run_tasks(cmd_args const& args)
         density.load();
         potential.generate(density, ctx->use_symmetry(), true);
         Band band(*ctx);
-        Hamiltonian0<double> H0(potential);
+        Hamiltonian0<double> H0(potential, true);
         if (!ctx->full_potential()) {
             band.initialize_subspace(ks, H0);
             if (ctx->hubbard_correction()) {
@@ -406,7 +436,7 @@ void run_tasks(cmd_args const& args)
                 //potential.U().calculate_hubbard_potential_and_energy(potential.U().occupation_matrix());
             }
         }
-        band.solve<double, double>(ks, H0, true, ctx->cfg().iterative_solver().energy_tolerance());
+        band.solve<double, double>(ks, H0, ctx->cfg().iterative_solver().energy_tolerance());
 
         ks.sync_band<double, sync_band_t::energy>();
         if (Communicator::world().rank() == 0) {
