@@ -1,4 +1,4 @@
-#include <sirius.h>
+#include <sirius.hpp>
 
 using namespace sirius;
 
@@ -9,8 +9,7 @@ void test_allgather()
 
     splindex<splindex_t::block> spl(N, Communicator::world().size(), Communicator::world().rank());
 
-    for (int i = 0; i < spl.local_size(); i++)
-    {
+    for (int i = 0; i < spl.local_size(); i++) {
         vec[spl[i]] = Communicator::world().rank() + 1.0;
     }
 
@@ -18,19 +17,17 @@ void test_allgather()
         sddk::pstdout pout(Communicator::world());
         if (Communicator::world().rank() == 0) pout.printf("before\n");
         pout.printf("rank : %i array : ", Communicator::world().rank());
-        for (int i = 0; i < N; i++)
-        {
+        for (int i = 0; i < N; i++) {
             pout.printf("%f ", vec[i]);
         }
         pout.printf("\n");
         pout.flush();
 
-        Communicator::world().allgather(&vec[0], spl.global_offset(), spl.local_size()); 
+        Communicator::world().allgather(&vec[0], spl.local_size(), spl.global_offset());
  
         if (Communicator::world().rank() == 0) pout.printf("after\n");
         pout.printf("rank : %i array : ", Communicator::world().rank());
-        for (int i = 0; i < N; i++)
-        {
+        for (int i = 0; i < N; i++) {
             pout.printf("%f ", vec[i]);
         }
         pout.printf("\n");
