@@ -1,4 +1,5 @@
-#include "test.hpp"
+#include "sirius.hpp"
+#include "testing.hpp"
 
 #ifdef __TEST_REAL
 typedef double gemm_type;
@@ -11,8 +12,6 @@ int const nop_gemm = 8;
 
 double test_gemm(int M, int N, int K, int transa)
 {
-    utils::timer t("test_gemm"); 
-    
     mdarray<gemm_type, 2> a, b, c;
     int imax, jmax;
     if (transa == 0) {
@@ -44,17 +43,20 @@ double test_gemm(int M, int N, int K, int transa)
     printf("a.ld() = %i\n", a.ld());
     printf("b.ld() = %i\n", b.ld());
     printf("c.ld() = %i\n", c.ld());
-    utils::timer t1("gemm_only"); 
-    linalg<device_t::CPU>::gemm(transa, 0, M, N, K, a.at(memory_t::host), a.ld(), b.at(memory_t::host), b.ld(), c.at(memory_t::host), c.ld());
-    double tval = t1.stop();
-    double perf = nop_gemm * 1e-9 * M * N * K / tval;
-    printf("execution time (sec) : %12.6f\n", tval);
+    const char ta[] = {'N', 'T', 'C'};
+    double t = -utils::wtime();
+    linalg(linalg_t::blas).gemm(ta[transa], 'N', M, N, K, &sddk::linalg_const<gemm_type>::one(),
+        a.at(memory_t::host), a.ld(), b.at(memory_t::host), b.ld(), &sddk::linalg_const<gemm_type>::zero(),
+        c.at(memory_t::host), c.ld());
+    t += utils::wtime();
+    double perf = nop_gemm * 1e-9 * M * N * K / t;
+    printf("execution time (sec) : %12.6f\n", t);
     printf("performance (GFlops) : %12.6f\n", perf);
 
     return perf;
 }
 
-#ifdef __SCALAPACK
+#ifdef SIRIUS_SCALAPACK
 double test_pgemm(int M, int N, int K, int nrow, int ncol, int transa, int n, int bs)
 {
     //== #ifdef _GPU_
@@ -94,18 +96,19 @@ double test_pgemm(int M, int N, int K, int nrow, int ncol, int transa, int n, in
         printf("testing parallel gemm with M, N, K = %i, %i, %i, opA = %i\n", M, N - n, K, transa);
         printf("nrow, ncol = %i, %i, bs = %i\n", nrow, ncol, bs);
     }
-    utils::timer t1("gemm_only");
+    double t = -utils::wtime();
     gemm_type one = 1;
     gemm_type zero = 0;
-    linalg<device_t::CPU>::gemm(transa, 0, M, N - n, K, one, a, 0, 0, b, 0, n, zero, c, 0, 0);
+    const char TA [] = {'N', 'T', 'C'};
+    linalg(linalg_t::scalapack).gemm(TA[transa], 'N', M, N - n, K, &one, a, 0, 0, b, 0, n, &zero, c, 0, 0);
     //== #ifdef _GPU_
     //== cuda_device_synchronize();
     //== #endif
-    double tval = t1.stop();
-    double perf = nop_gemm * 1e-9 * M * (N - n) * K / tval / nrow / ncol;
+    t += utils::wtime();
+    double perf = nop_gemm * 1e-9 * M * (N - n) * K / t / nrow / ncol;
     if (Communicator::world().rank() == 0)
     {
-        printf("execution time : %12.6f seconds\n", tval);
+        printf("execution time : %12.6f seconds\n", t);
         printf("performance    : %12.6f GFlops / rank\n", perf);
     }
 
@@ -193,22 +196,19 @@ int main(int argn, char **argv)
 
     sirius::initialize(true);
 
-    if (nrow * ncol == 1)
-    {
-        Measurement perf;
+    if (nrow * ncol == 1) {
+        sirius::Measurement perf;
         for (int i = 0; i < repeat; i++) {
             perf.push_back(test_gemm(M, N, K, transa));
         }
         printf("average performance: %12.6f GFlops / rank,  sigma: %12.6f\n", perf.average(), perf.sigma());
     }
-    else
-    {
-        #ifdef __SCALAPACK
+    else {
+#ifdef SIRIUS_SCALAPACK
         int n = args.value<int>("n", 0);
         int bs = args.value<int>("bs");
         double perf = 0;
-        for (int i = 0; i < repeat; i++) 
-        {
+        for (int i = 0; i < repeat; i++) {
             perf += test_pgemm(M, N, K, nrow, ncol, transa, n, bs);
             //if (M != N)
             //{
@@ -223,9 +223,9 @@ int main(int argn, char **argv)
             printf("\n");
             printf("average performance    : %12.6f GFlops / rank\n", perf / repeat);
         }
-        #else
-        TERMINATE_NO_SCALAPACK;
-        #endif
+#else
+        throw std::runtime_error("not compiled with ScaLAPACK");
+#endif
     }
 
     sirius::finalize();
