@@ -1,10 +1,10 @@
-#include <sirius.h>
+#include <sirius.hpp>
 
 using namespace sirius;
 
 int test1(int size)
 {
-    utils::timer t1("test1");
+    double t = -utils::wtime();
     sddk::mdarray<char, 1> buf(size);
     sddk::mdarray<char, 1> buf1(size);
     buf.zero();
@@ -20,7 +20,7 @@ int test1(int size)
         Communicator::world().recv(buf1.at(memory_t::host), size, rank2, Communicator::get_tag(rank, rank2));
         req.wait();
     }
-    double t = t1.stop();
+    t += utils::wtime();
     if (Communicator::world().rank() == 0) {
         printf("time : %f sec.\n", t);
     }
