@@ -40,9 +40,9 @@ dmatrix<T>::dmatrix(int num_rows__, int num_cols__, BLACS_grid const& blacs_grid
     , blacs_grid_(&blacs_grid__)
     , spl_row_(num_rows_, blacs_grid__.num_ranks_row(), blacs_grid__.rank_row(), bs_row_)
     , spl_col_(num_cols_, blacs_grid__.num_ranks_col(), blacs_grid__.rank_col(), bs_col_)
-    , spla_distri_(spla::MatrixDistribution::create_blacs_block_cyclic_from_mapping(
-          blacs_grid__.comm().mpi_comm(), blacs_grid__.rank_map().data(), blacs_grid__.num_ranks_row(), blacs_grid__.num_ranks_col(), bs_row__,
-          bs_col__))
+    , spla_dist_(spla::MatrixDistribution::create_blacs_block_cyclic_from_mapping(
+          blacs_grid__.comm().mpi_comm(), blacs_grid__.rank_map().data(), blacs_grid__.num_ranks_row(),
+          blacs_grid__.num_ranks_col(), bs_row__,bs_col__))
 {
     init();
 }
@@ -62,9 +62,9 @@ dmatrix<T>::dmatrix(T* ptr__, int num_rows__, int num_cols__, BLACS_grid const&
     , blacs_grid_(&blacs_grid__)
     , spl_row_(num_rows_, blacs_grid__.num_ranks_row(), blacs_grid__.rank_row(), bs_row_)
     , spl_col_(num_cols_, blacs_grid__.num_ranks_col(), blacs_grid__.rank_col(), bs_col_)
-    , spla_distri_(spla::MatrixDistribution::create_blacs_block_cyclic_from_mapping(
-          blacs_grid__.comm().mpi_comm(), blacs_grid__.rank_map().data(), blacs_grid__.num_ranks_row(), blacs_grid__.num_ranks_col(), bs_row__,
-          bs_col__))
+    , spla_dist_(spla::MatrixDistribution::create_blacs_block_cyclic_from_mapping(
+          blacs_grid__.comm().mpi_comm(), blacs_grid__.rank_map().data(), blacs_grid__.num_ranks_row(),
+          blacs_grid__.num_ranks_col(), bs_row__, bs_col__))
 {
     init();
 }
@@ -121,9 +121,9 @@ dmatrix<T>::dmatrix(int num_rows__, int num_cols__, BLACS_grid const& blacs_grid
     , blacs_grid_(&blacs_grid__)
     , spl_row_(num_rows_, blacs_grid__.num_ranks_row(), blacs_grid__.rank_row(), bs_row_)
     , spl_col_(num_cols_, blacs_grid__.num_ranks_col(), blacs_grid__.rank_col(), bs_col_)
-    , spla_distri_(spla::MatrixDistribution::create_blacs_block_cyclic_from_mapping(
-          blacs_grid__.comm().mpi_comm(), blacs_grid__.rank_map().data(), blacs_grid__.num_ranks_row(), blacs_grid__.num_ranks_col(), bs_row__,
-          bs_col__))
+    , spla_dist_(spla::MatrixDistribution::create_blacs_block_cyclic_from_mapping(
+          blacs_grid__.comm().mpi_comm(), blacs_grid__.rank_map().data(), blacs_grid__.num_ranks_row(),
+          blacs_grid__.num_ranks_col(), bs_row__, bs_col__))
 {
     init();
 }
