@@ -128,6 +128,46 @@ Wave_functions<T>::Wave_functions(const Gvec_partition& gkvecp__, int num_atoms_
     }
 }
 
+template <typename T>
+Wave_functions<T>::Wave_functions(memory_pool &mp__, const Gvec_partition& gkvecp__, int num_atoms__, std::function<int(int)> mt_size__,
+                               int num_wf__, memory_t preferred_memory_t__, int num_sc__)
+    : comm_(gkvecp__.gvec().comm())
+    , gkvecp_(gkvecp__)
+    , num_wf_(num_wf__)
+    , num_sc_(num_sc__)
+    , has_mt_(true)
+    , preferred_memory_t_(preferred_memory_t__)
+{
+    if (!(num_sc__ == 1 || num_sc__ == 2)) {
+        TERMINATE("wrong number of spin components");
+    }
+
+    for (int ispn = 0; ispn < num_sc_; ispn++) {
+        pw_coeffs_[ispn] = std::unique_ptr<matrix_storage<std::complex<T>, matrix_storage_t::slab>>(
+            new matrix_storage<std::complex<T>, matrix_storage_t::slab>(mp__, gkvecp_, num_wf_));
+    }
+
+    /* TODO: remove reduntant code */
+    spl_num_atoms_   = splindex<splindex_t::block>(num_atoms__, comm_.size(), comm_.rank());
+    mt_coeffs_distr_ = block_data_descriptor(comm_.size());
+
+    for (int ia = 0; ia < num_atoms__; ia++) {
+        int rank = spl_num_atoms_.local_rank(ia);
+        if (rank == comm_.rank()) {
+            offset_mt_coeffs_.push_back(mt_coeffs_distr_.counts[rank]);
+        }
+        mt_coeffs_distr_.counts[rank] += mt_size__(ia);
+    }
+    mt_coeffs_distr_.calc_offsets();
+
+    num_mt_coeffs_ = mt_coeffs_distr_.offsets.back() + mt_coeffs_distr_.counts.back();
+
+    for (int ispn = 0; ispn < num_sc_; ispn++) {
+        mt_coeffs_[ispn] = std::unique_ptr<matrix_storage<std::complex<T>, matrix_storage_t::slab>>(
+            new matrix_storage<std::complex<T>, matrix_storage_t::slab>(mp__, mt_coeffs_distr_.counts[comm_.rank()], num_wf_));
+    }
+}
+
 template <typename T>
 void Wave_functions<T>::copy_from(device_t pu__, int n__, const Wave_functions<T>& src__, int ispn__, int i0__,
                                   int jspn__, int j0__)
@@ -136,7 +176,8 @@ void Wave_functions<T>::copy_from(device_t pu__, int n__, const Wave_functions<T
     assert(jspn__ == 0 || jspn__ == 1);
 
     int ngv = pw_coeffs(jspn__).num_rows_loc();
-    int nmt = has_mt() ? mt_coeffs(jspn__).num_rows_loc() : 0;
+    bool has_mt = this->has_mt() && src__.has_mt();
+    int nmt = has_mt ? mt_coeffs(jspn__).num_rows_loc() : 0;
 
     switch (pu__) {
         case device_t::CPU: {
@@ -145,7 +186,7 @@ void Wave_functions<T>::copy_from(device_t pu__, int n__, const Wave_functions<T
                       src__.pw_coeffs(ispn__).prime().at(memory_t::host, 0, i0__) + ngv * n__,
                       pw_coeffs(jspn__).prime().at(memory_t::host, 0, j0__));
             /* copy MT part */
-            if (has_mt()) {
+            if (has_mt) {
                 std::copy(src__.mt_coeffs(ispn__).prime().at(memory_t::host, 0, i0__),
                           src__.mt_coeffs(ispn__).prime().at(memory_t::host, 0, i0__) + nmt * n__,
                           mt_coeffs(jspn__).prime().at(memory_t::host, 0, j0__));
@@ -157,7 +198,7 @@ void Wave_functions<T>::copy_from(device_t pu__, int n__, const Wave_functions<T
             acc::copy(pw_coeffs(jspn__).prime().at(memory_t::device, 0, j0__),
                       src__.pw_coeffs(ispn__).prime().at(memory_t::device, 0, i0__), ngv * n__);
             /* copy MT part */
-            if (has_mt()) {
+            if (has_mt) {
                 acc::copy(mt_coeffs(jspn__).prime().at(memory_t::device, 0, j0__),
                           src__.mt_coeffs(ispn__).prime().at(memory_t::device, 0, i0__), nmt * n__);
             }
@@ -172,12 +213,14 @@ void Wave_functions<T>::copy_from(const Wave_functions<T>& src__, int n__, int i
     assert(ispn__ == 0 || ispn__ == 1);
     assert(jspn__ == 0 || jspn__ == 1);
 
+    bool has_mt = this->has_mt() && src__.has_mt();
+
     int ngv = pw_coeffs(jspn__).num_rows_loc();
-    int nmt = has_mt() ? mt_coeffs(jspn__).num_rows_loc() : 0;
+    int nmt = has_mt ? mt_coeffs(jspn__).num_rows_loc() : 0;
 
     copy(src__.preferred_memory_t(), src__.pw_coeffs(ispn__).prime().at(src__.preferred_memory_t(), 0, i0__),
          preferred_memory_t(), pw_coeffs(jspn__).prime().at(preferred_memory_t(), 0, j0__), ngv * n__);
-    if (has_mt()) {
+    if (has_mt) {
         copy(src__.preferred_memory_t(), src__.mt_coeffs(ispn__).prime().at(src__.preferred_memory_t(), 0, i0__),
              preferred_memory_t(), mt_coeffs(jspn__).prime().at(preferred_memory_t(), 0, j0__), nmt * n__);
     }
@@ -186,8 +229,10 @@ void Wave_functions<T>::copy_from(const Wave_functions<T>& src__, int n__, int i
 template <typename T>
 std::complex<T> Wave_functions<T>::checksum_pw(device_t pu__, int ispn__, int i0__, int n__) const
 {
-    assert(n__ != 0);
     std::complex<T> cs(0, 0);
+    if (n__ == 0) {
+        return cs;
+    }
     for (int s = s0(ispn__); s <= s1(ispn__); s++) {
         cs += pw_coeffs(s).checksum(pu__, i0__, n__);
     }
@@ -198,9 +243,8 @@ std::complex<T> Wave_functions<T>::checksum_pw(device_t pu__, int ispn__, int i0
 template <typename T>
 std::complex<T> Wave_functions<T>::checksum_mt(device_t pu__, int ispn__, int i0__, int n__) const
 {
-    assert(n__ != 0);
     std::complex<T> cs(0, 0);
-    if (!this->has_mt_) {
+    if (!this->has_mt_ || n__ == 0) {
         return cs;
     }
     for (int s = s0(ispn__); s <= s1(ispn__); s++) {
@@ -212,24 +256,6 @@ std::complex<T> Wave_functions<T>::checksum_mt(device_t pu__, int ispn__, int i0
     return cs;
 }
 
-template <typename T>
-void Wave_functions<T>::print_checksum(device_t pu__, std::string label__, int N__, int n__) const
-{
-    for (int ispn = 0; ispn < num_sc(); ispn++) {
-        auto cs1 = this->checksum_pw(pu__, ispn, N__, n__);
-        auto cs2 = this->checksum_mt(pu__, ispn, N__, n__);
-        if (this->comm().rank() == 0) {
-            std::stringstream s;
-            s << ispn;
-            utils::print_checksum(label__ + "_pw_" + s.str(), cs1);
-            if (this->has_mt_) {
-                utils::print_checksum(label__ + "_mt_" + s.str(), cs2);
-            }
-            utils::print_checksum(label__ + "_" + s.str(), cs1 + cs2);
-        }
-    }
-}
-
 template <typename T>
 void Wave_functions<T>::zero_pw(device_t pu__, int ispn__, int i0__, int n__) // TODO: pass memory_t
 {
@@ -380,6 +406,178 @@ void Wave_functions<T>::copy_to(spin_range spins__, memory_t mem__, int i0__, in
     }
 }
 
+template <typename T>
+mdarray<std::complex<T>, 1> Wave_functions<T>::dot(device_t pu__, spin_range spins__, Wave_functions<T> const &phi, int n__) const
+{
+    mdarray<std::complex<T>, 1> s(n__, memory_t::host, "dot");
+    s.zero();
+
+    for (int is : spins__) {
+        switch (pu__) {
+            case device_t::CPU: {
+                #pragma omp parallel for
+                for (int i = 0; i < n__; i++) {
+                    for (int ig = 0; ig < pw_coeffs(is).num_rows_loc(); ig++) {
+                        auto x = pw_coeffs(is).prime(ig, i);
+                        auto y = phi.pw_coeffs(is).prime(ig, i);
+                        s[i] += std::conj(x) * y;
+                    }
+                    // todo, do something here.
+                    // if (gkvecp_.gvec().reduced()) {
+                    //     if (comm_.rank() == 0) {
+                    //         s[i] = 2 * s[i] - std::pow(pw_coeffs(is).prime(0, i).real(), 2);
+                    //     } else {
+                    //         s[i] *= 2;
+                    //     }
+                    // }
+                    if (has_mt()) {
+                        for (int j = 0; j < mt_coeffs(is).num_rows_loc(); j++) {
+                            auto x = mt_coeffs(is).prime(j, i);
+                            auto y = phi.mt_coeffs(is).prime(j, i);
+                            s[i] += std::conj(x) * y;
+                        }
+                    }
+                }
+                break;
+            }
+            case device_t::GPU: {
+                throw "not implemented yet";
+            }
+        }
+    }
+
+    if (pu__ == device_t::GPU) {
+        s.copy_to(memory_t::host);
+    }
+
+    comm_.allreduce(s.at(memory_t::host), n__);
+
+    return s;
+}
+
+template <typename T>
+template <typename Ta>
+void Wave_functions<T>::axpby(device_t pu__, spin_range spins__, Ta alpha, Wave_functions<T> const &phi, Ta beta, int n__)
+{
+    for (int is : spins__) {
+        switch (pu__) {
+            case device_t::CPU: {
+                #pragma omp parallel for
+                for (int i = 0; i < n__; i++) {
+                    for (int ig = 0; ig < pw_coeffs(is).num_rows_loc(); ig++) {
+                        auto x = phi.pw_coeffs(is).prime(ig, i);
+                        auto y = pw_coeffs(is).prime(ig, i);
+                        pw_coeffs(is).prime(ig, i) = alpha * x + beta * y;
+                    }
+                    if (has_mt()) {
+                        for (int j = 0; j < mt_coeffs(is).num_rows_loc(); j++) {
+                            auto x = phi.mt_coeffs(is).prime(j, i);
+                            auto y = mt_coeffs(is).prime(j, i);
+                            mt_coeffs(is).prime(j, i) = alpha * x + beta * y;
+                        }
+                    }
+                }
+                break;
+            }
+            case device_t::GPU: {
+                throw "not implemented yet";
+            }
+        }
+    }
+}
+
+template <typename T>
+template <typename Ta>
+void Wave_functions<T>::xpby(device_t pu__, spin_range spins__, Wave_functions<T> const &phi, std::vector<Ta> const &betas, int n__)
+{
+    for (int is : spins__) {
+        switch (pu__) {
+            case device_t::CPU: {
+                #pragma omp parallel for
+                for (int i = 0; i < n__; i++) {
+                    auto beta = betas[i];
+
+                    for (int ig = 0; ig < pw_coeffs(is).num_rows_loc(); ig++) {
+                        auto x = phi.pw_coeffs(is).prime(ig, i);
+                        auto y = pw_coeffs(is).prime(ig, i);
+                        pw_coeffs(is).prime(ig, i) = x + beta * y;
+                    }
+                    if (has_mt()) {
+                        for (int j = 0; j < mt_coeffs(is).num_rows_loc(); j++) {
+                            auto x = phi.mt_coeffs(is).prime(j, i);
+                            auto y = mt_coeffs(is).prime(j, i);
+                            mt_coeffs(is).prime(j, i) = x + beta * y;
+                        }
+                    }
+                }
+                break;
+            }
+            case device_t::GPU: {
+                throw "not implemented yet";
+            }
+        }
+    }
+}
+
+template <typename T>
+template <typename Ta>
+void Wave_functions<T>::axpy(device_t pu__, spin_range spins__, std::vector<Ta> const &alphas, Wave_functions<T> const &phi, int n__)
+{
+    for (int is : spins__) {
+        switch (pu__) {
+            case device_t::CPU: {
+                #pragma omp parallel for
+                for (int i = 0; i < n__; i++) {
+                    auto alpha = alphas[i];
+
+                    for (int ig = 0; ig < pw_coeffs(is).num_rows_loc(); ig++) {
+                        pw_coeffs(is).prime(ig, i) += alpha * phi.pw_coeffs(is).prime(ig, i);
+                    }
+                    if (has_mt()) {
+                        for (int j = 0; j < mt_coeffs(is).num_rows_loc(); j++) {
+                            mt_coeffs(is).prime(j, i) += alpha * phi.mt_coeffs(is).prime(j, i);
+                        }
+                    }
+                }
+                break;
+            }
+            case device_t::GPU: {
+                throw "not implemented yet";
+            }
+        }
+    }
+}
+
+template <typename T>
+template <typename Ta>
+void Wave_functions<T>::axpy_scatter(device_t pu__, spin_range spins__, std::vector<Ta> const &alphas, Wave_functions<T> const &phi, std::vector<size_t> const &ids, int n__)
+{
+    for (int is : spins__) {
+        switch (pu__) {
+            case device_t::CPU: {
+                #pragma omp parallel for
+                for (int i = 0; i < n__; i++) {
+                    auto ii = ids[i];
+                    auto alpha = alphas[i];
+
+                    for (int ig = 0; ig < pw_coeffs(is).num_rows_loc(); ig++) {
+                        pw_coeffs(is).prime(ig, ii) += alpha * phi.pw_coeffs(is).prime(ig, i);
+                    }
+                    if (has_mt()) {
+                        for (int j = 0; j < mt_coeffs(is).num_rows_loc(); j++) {
+                            mt_coeffs(is).prime(j, ii) += alpha * phi.mt_coeffs(is).prime(j, i);
+                        }
+                    }
+                }
+                break;
+            }
+            case device_t::GPU: {
+                throw "not implemented yet";
+            }
+        }
+    }
+}
+
 template <typename T>
 mdarray<T, 1> Wave_functions<T>::sumsqr(device_t pu__, spin_range spins__, int n__) const
 {
@@ -443,4 +641,19 @@ template class Wave_functions<double>;
 #ifdef USE_FP32
 template class Wave_functions<float>;
 #endif
+
+template void Wave_functions<double>::axpby(device_t pu__, spin_range spins__, double alpha, Wave_functions<double> const &phi, double beta, int n__);
+template void Wave_functions<double>::axpby(device_t pu__, spin_range spins__, double_complex alpha, Wave_functions<double> const &phi, double_complex beta, int n__);
+
+template void Wave_functions<double>::xpby(device_t pu__, spin_range spins__, Wave_functions<double> const &phi, std::vector<double> const &betas, int n__);
+template void Wave_functions<double>::xpby(device_t pu__, spin_range spins__, Wave_functions<double> const &phi, std::vector<double_complex> const &betas, int n__);
+
+template void Wave_functions<double>::axpy(device_t pu__, spin_range spins__, std::vector<double> const &alphas, Wave_functions<double> const &phi, int n__);
+template void Wave_functions<double>::axpy(device_t pu__, spin_range spins__, std::vector<double_complex> const &alphas, Wave_functions<double> const &phi, int n__);
+
+template void Wave_functions<double>::axpy_scatter(device_t pu__, spin_range spins__, std::vector<double_complex> const &alphas, Wave_functions<double> const &phi, std::vector<size_t> const &ids, int n__);
+template void Wave_functions<double>::axpy_scatter(device_t pu__, spin_range spins__, std::vector<double> const &alphas, Wave_functions<double> const &phi, std::vector<size_t> const &ids, int n__);
+
 } // namespace sddk
+
+
