@@ -68,7 +68,6 @@ double unit_step_function_form_factors(double R__, double g__);
 class Simulation_context : public Simulation_parameters
 {
   private:
-
     /// Communicator for this simulation.
     Communicator const& comm_;
 
@@ -179,16 +178,10 @@ class Simulation_context : public Simulation_parameters
     std::function<void(int, double, double*, int, int)> aug_ri_djl_callback_{nullptr};
 
     /// Radial integrals of atomic wave-functions.
-    std::unique_ptr<Radial_integrals_atomic_wf<false>> atomic_wf_ri_;
+    std::unique_ptr<Radial_integrals_atomic_wf<false>> ps_atomic_wf_ri_;
 
     /// Radial integrals of atomic wave-functions with derivatives of spherical Bessel functions.
-    std::unique_ptr<Radial_integrals_atomic_wf<true>> atomic_wf_ri_djl_;
-
-    /// Radial integrals of hubbard wave-functions.
-    std::unique_ptr<Radial_integrals_atomic_wf<false>> hubbard_wf_ri_;
-
-    /// Radial integrals of hubbard wave-functions with derivatives of spherical Bessel functions.
-    std::unique_ptr<Radial_integrals_atomic_wf<true>> hubbard_wf_ri_djl_;
+    std::unique_ptr<Radial_integrals_atomic_wf<true>> ps_atomic_wf_ri_djl_;
 
     /// Radial integrals of pseudo-core charge density.
     std::unique_ptr<Radial_integrals_rho_core_pseudo<false>> ps_core_ri_;
@@ -200,6 +193,8 @@ class Simulation_context : public Simulation_parameters
     std::function<void(int, int, double*, double*)> rhoc_ri_djl_callback_{nullptr};
 
     std::function<void(int, int, double*, double*)> ps_rho_ri_callback_{nullptr};
+    std::function<void(int, double, double*, int)> ps_atomic_wf_ri_callback_{nullptr};
+    std::function<void(int, double, double*, int)> ps_atomic_wf_ri_djl_callback_{nullptr};
 
     /// Radial integrals of total pseudo-charge density.
     std::unique_ptr<Radial_integrals_rho_pseudo> ps_rho_ri_;
@@ -252,9 +247,11 @@ class Simulation_context : public Simulation_parameters
     /// Callback function to compute effective potential.
     std::function<void(void)> veff_callback_{nullptr};
 
-    // Spla context
+    /// Spla context.
     std::shared_ptr<::spla::Context> spla_ctx_{new ::spla::Context{SPLA_PU_HOST}};
 
+    std::shared_ptr<std::ostream> output_stream_;
+
     mutable double evp_work_count_{0};
     mutable int num_loc_op_applied_{0};
     /// Total number of iterative solver steps.
@@ -320,7 +317,6 @@ class Simulation_context : public Simulation_parameters
     Simulation_context(Simulation_context const&) = delete;
 
   public:
-
     /// Create an empty simulation context with an explicit communicator.
     Simulation_context(Communicator const& comm__ = Communicator::world())
         : comm_(comm__)
@@ -338,7 +334,6 @@ class Simulation_context : public Simulation_parameters
         start();
     }
 
-
     /// Create a simulation context with world communicator and load parameters from JSON string or JSON file.
     Simulation_context(std::string const& str__)
         : comm_(Communicator::world())
@@ -350,7 +345,7 @@ class Simulation_context : public Simulation_parameters
     }
 
     // /// Create a simulation context with world communicator and load parameters from JSON string or JSON file.
-    Simulation_context(std::string const& str__, Communicator const &comm__)
+    Simulation_context(std::string const& str__, Communicator const& comm__)
         : comm_(comm__)
     {
         unit_cell_ = std::make_unique<Unit_cell>(*this, comm_);
@@ -391,7 +386,7 @@ class Simulation_context : public Simulation_parameters
     {
         if (this->comm().rank() == 0 && this->cfg().control().verbosity() >= level__) {
             auto strings = ::rte::split(s.str());
-            for (auto& e: strings) {
+            for (auto& e : strings) {
                 std::cout << "[" << label__ << "] " << e << std::endl;
             }
         }
@@ -509,22 +504,22 @@ class Simulation_context : public Simulation_parameters
 
     inline Eigensolver& std_evp_solver()
     {
-        return* std_evp_solver_;
+        return *std_evp_solver_;
     }
 
     inline Eigensolver const& std_evp_solver() const
     {
-        return* std_evp_solver_;
+        return *std_evp_solver_;
     }
 
     inline Eigensolver& gen_evp_solver()
     {
-        return* gen_evp_solver_;
+        return *gen_evp_solver_;
     }
 
     inline Eigensolver const& gen_evp_solver() const
     {
-        return* gen_evp_solver_;
+        return *gen_evp_solver_;
     }
 
     /// Phase factors \f$ e^{i {\bf G} {\bf r}_{\alpha}} \f$
@@ -536,7 +531,7 @@ class Simulation_context : public Simulation_parameters
     /// Phase factors \f$ e^{i {\bf G} {\bf r}_{\alpha}} \f$
     inline double_complex gvec_phase_factor(int ig__, int ia__) const
     {
-        return gvec_phase_factor(gvec().gvec(ig__), ia__);
+        return gvec_phase_factor(gvec().gvec<index_domain_t::global>(ig__), ia__);
     }
 
     inline sddk::mdarray<int, 2> const& gvec_coord() const
@@ -555,8 +550,7 @@ class Simulation_context : public Simulation_parameters
     /// Make periodic function out of form factors.
     /** Return vector of plane-wave coefficients */ // TODO: return mdarray
     template <index_domain_t index_domain, typename F>
-    inline std::vector<double_complex>
-    make_periodic_function(F&& form_factors__) const
+    inline std::vector<double_complex> make_periodic_function(F&& form_factors__) const
     {
         PROFILE("sirius::Simulation_context::make_periodic_function");
 
@@ -569,7 +563,7 @@ class Simulation_context : public Simulation_parameters
         for (int igloc = 0; igloc < gvec().count(); igloc++) {
             /* global index of G-vector */
             int ig   = gvec().offset() + igloc;
-            double g = gvec().gvec_len(ig);
+            double g = gvec().gvec_len<index_domain_t::local>(igloc);
 
             int j = (index_domain == index_domain_t::local) ? igloc : ig;
             for (int iat = 0; iat < unit_cell().num_atom_types(); iat++) {
@@ -586,8 +580,7 @@ class Simulation_context : public Simulation_parameters
 
     /// Make periodic out of form factors computed for G-shells.
     template <index_domain_t index_domain>
-    inline std::vector<double_complex>
-    make_periodic_function(sddk::mdarray<double, 2>& form_factors__) const
+    inline std::vector<double_complex> make_periodic_function(sddk::mdarray<double, 2>& form_factors__) const
     {
         PROFILE("sirius::Simulation_context::make_periodic_function");
 
@@ -652,24 +645,14 @@ class Simulation_context : public Simulation_parameters
         return *aug_ri_djl_;
     }
 
-    inline auto const& atomic_wf_ri() const
+    inline auto const& ps_atomic_wf_ri() const
     {
-        return *atomic_wf_ri_;
+        return *ps_atomic_wf_ri_;
     }
 
-    inline auto const& atomic_wf_djl() const
+    inline auto const& ps_atomic_wf_ri_djl() const
     {
-        return *atomic_wf_ri_djl_;
-    }
-
-    inline auto const& hubbard_wf_ri() const
-    {
-        return *hubbard_wf_ri_;
-    }
-
-    inline auto const& hubbard_wf_djl() const
-    {
-        return *hubbard_wf_ri_djl_;
+        return *ps_atomic_wf_ri_djl_;
     }
 
     inline auto const& ps_core_ri() const
@@ -873,6 +856,16 @@ class Simulation_context : public Simulation_parameters
         rhoc_ri_djl_callback_ = fptr__;
     }
 
+    inline void ps_atomic_wf_ri_callback(void (*fptr__)(int, double, double*, int))
+    {
+        ps_atomic_wf_ri_callback_ = fptr__;
+    }
+
+    inline void ps_atomic_wf_ri_djl_callback(void (*fptr__)(int, double, double*, int))
+    {
+        ps_atomic_wf_ri_djl_callback_ = fptr__;
+    }
+
     /// Set callback function to compute band occupations
     inline void band_occ_callback(void (*fptr__)(void))
     {
@@ -901,14 +894,24 @@ class Simulation_context : public Simulation_parameters
         dict["config"] = cfg().dict();
         bool const cart_pos{false};
         dict["config"]["unit_cell"] = unit_cell().serialize(cart_pos);
-        auto fftgrid = {spfft_transform_coarse_->dim_x(), spfft_transform_coarse_->dim_y(), spfft_transform_coarse_->dim_z()};
-        dict["fft_coarse_grid"] = fftgrid;
-        dict["mpi_grid"] = mpi_grid_dims();
-        dict["omega"] = unit_cell().omega();
-        dict["chemical_formula"] = unit_cell().chemical_formula();
-        dict["num_atoms"] = unit_cell().num_atoms();
+        auto fftgrid                = {spfft_transform_coarse_->dim_x(), spfft_transform_coarse_->dim_y(),
+                        spfft_transform_coarse_->dim_z()};
+        dict["fft_coarse_grid"]     = fftgrid;
+        dict["mpi_grid"]            = mpi_grid_dims();
+        dict["omega"]               = unit_cell().omega();
+        dict["chemical_formula"]    = unit_cell().chemical_formula();
+        dict["num_atoms"]           = unit_cell().num_atoms();
         return dict;
     }
+
+    std::ostream& out() const
+    {
+        if (output_stream_ == nullptr) {
+            return std::cout;
+        } else {
+            return *output_stream_;
+        }
+    }
 };
 
 } // namespace sirius
