@@ -28,8 +28,11 @@
 
 namespace sirius {
 
-void DFT_ground_state::initial_state()
+void
+DFT_ground_state::initial_state()
 {
+    PROFILE("sirius::DFT_ground_state::initial_state");
+
     density_.initial_density();
     potential_.generate(density_, ctx_.use_symmetry(), true);
     if (!ctx_.full_potential()) {
@@ -42,13 +45,14 @@ void DFT_ground_state::initial_state()
 #endif
 
         } else {
-            Hamiltonian0<double> H0(potential_);
+            Hamiltonian0<double> H0(potential_, true);
             Band(ctx_).initialize_subspace(kset_, H0);
         }
     }
 }
 
-void DFT_ground_state::update()
+void
+DFT_ground_state::update()
 {
     PROFILE("sirius::DFT_ground_state::update");
 
@@ -62,12 +66,13 @@ void DFT_ground_state::update()
     }
 }
 
-double DFT_ground_state::energy_kin_sum_pw() const
+double
+DFT_ground_state::energy_kin_sum_pw() const
 {
     double ekin{0};
 
     for (int ikloc = 0; ikloc < kset_.spl_num_kpoints().local_size(); ikloc++) {
-        int ik = kset_.spl_num_kpoints(ikloc);
+        int ik  = kset_.spl_num_kpoints(ikloc);
         auto kp = kset_.get<double>(ik);
 
         #pragma omp parallel for schedule(static) reduction(+:ekin)
@@ -78,7 +83,7 @@ double DFT_ground_state::energy_kin_sum_pw() const
             for (int ispin = 0; ispin < ctx_.num_spins(); ispin++) {
                 for (int i = 0; i < kp->num_occupied_bands(ispin); i++) {
                     double f = kp->band_occupancy(i, ispin);
-                    auto z = kp->spinor_wave_functions().pw_coeffs(ispin).prime(igloc, i);
+                    auto z   = kp->spinor_wave_functions().pw_coeffs(ispin).prime(igloc, i);
                     d += f * (std::pow(z.real(), 2) + std::pow(z.imag(), 2));
                 }
             }
@@ -87,17 +92,19 @@ double DFT_ground_state::energy_kin_sum_pw() const
             }
             ekin += 0.5 * d * kp->weight() * Gk.length2();
         } // igloc
-    } // ikloc
+    }     // ikloc
     ctx_.comm().allreduce(&ekin, 1);
     return ekin;
 }
 
-double DFT_ground_state::total_energy() const
+double
+DFT_ground_state::total_energy() const
 {
     return sirius::total_energy(ctx_, kset_, density_, potential_, ewald_energy_) + this->scf_energy_;
 }
 
-json DFT_ground_state::serialize()
+json
+DFT_ground_state::serialize()
 {
     nlohmann::json dict;
 
@@ -114,19 +121,20 @@ json DFT_ground_state::serialize()
     dict["energy"]["kin"]           = energy_kin(ctx_, kset_, density_, potential_);
     dict["energy"]["ewald"]         = ewald_energy_;
     if (!ctx_.full_potential()) {
-        dict["energy"]["vloc"]      = energy_vloc(density_, potential_);
+        dict["energy"]["vloc"] = energy_vloc(density_, potential_);
     }
     dict["energy"]["scf_correction"] = this->scf_energy_;
-    dict["energy"]["entropy_sum"]   = kset_.entropy_sum();
-    dict["efermi"]                  = kset_.energy_fermi();
-    dict["band_gap"]                = kset_.band_gap();
-    dict["core_leakage"]            = density_.core_leakage();
+    dict["energy"]["entropy_sum"]    = kset_.entropy_sum();
+    dict["efermi"]                   = kset_.energy_fermi();
+    dict["band_gap"]                 = kset_.band_gap();
+    dict["core_leakage"]             = density_.core_leakage();
 
     return dict;
 }
 
 /// A quick check of self-constent density in case of pseudopotential.
-json DFT_ground_state::check_scf_density()
+json
+DFT_ground_state::check_scf_density()
 {
     if (ctx_.full_potential()) {
         return json();
@@ -143,11 +151,11 @@ json DFT_ground_state::check_scf_density()
     /* generate potential from existing density */
     pot.generate(density_, ctx_.use_symmetry(), true);
     /* create new Hamiltonian */
-    Hamiltonian0<double> H0(pot);
+    Hamiltonian0<double> H0(pot, true);
     /* initialize the subspace */
     Band(ctx_).initialize_subspace(kset_, H0);
     /* find new wave-functions */
-    Band(ctx_).solve<double, double>(kset_, H0, true, ctx_.cfg().settings().itsol_tol_min());
+    Band(ctx_).solve<double, double>(kset_, H0, ctx_.cfg().settings().itsol_tol_min());
     /* find band occupancies */
     kset_.find_band_occupancies<double>();
     /* generate new density from the occupied wave-functions */
@@ -170,7 +178,9 @@ json DFT_ground_state::check_scf_density()
     return dict;
 }
 
-json DFT_ground_state::find(double density_tol, double energy_tol, double itsol_tol, int num_dft_iter, bool write_state)
+json
+DFT_ground_state::find(double density_tol__, double energy_tol__, double iter_solver_tol__, int num_dft_iter__,
+                       bool write_state__)
 {
     PROFILE("sirius::DFT_ground_state::scf_loop");
 
@@ -187,30 +197,30 @@ json DFT_ground_state::find(double density_tol, double energy_tol, double itsol_
     Density rho1(ctx_);
 
     std::stringstream s;
-    s << "density_tol            : " << density_tol << std::endl
-      << "energy_tol             : " << energy_tol << std::endl
-      << "itsol_tol (initial)    : " << itsol_tol << std::endl
-      << "itsol_tol_min          : " << ctx_.cfg().settings().itsol_tol_min() << std::endl
-      << "num_dft_iter           : " << num_dft_iter;
+    s << "density_tol               : " << density_tol__ << std::endl
+      << "energy_tol                : " << energy_tol__ << std::endl
+      << "iter_solver_tol (initial) : " << iter_solver_tol__ << std::endl
+      << "iter_solver_tol (target)  : " << ctx_.cfg().settings().itsol_tol_min() << std::endl
+      << "num_dft_iter              : " << num_dft_iter__;
     ctx_.message(1, __func__, s);
 
-    for (int iter = 0; iter < num_dft_iter; iter++) {
+    for (int iter = 0; iter < num_dft_iter__; iter++) {
         PROFILE("sirius::DFT_ground_state::scf_loop|iteration");
         std::stringstream s;
         s << std::endl;
         s << "+------------------------------+" << std::endl
-          << "| SCF iteration " << std::setw(3) << iter << " out of " << std::setw(3) << num_dft_iter << '|' << std::endl
+          << "| SCF iteration " << std::setw(3) << iter << " out of " << std::setw(3) << num_dft_iter__ << '|' << std::endl
           << "+------------------------------+" << std::endl;
         ctx_.message(2, __func__, s);
 
         if (ctx_.cfg().parameters().precision_wf() == "fp32") {
 #if defined(USE_FP32)
-            Hamiltonian0<float> H0(potential_);
+            Hamiltonian0<float> H0(potential_, true);
             /* find new wave-functions */
             if (ctx_.cfg().parameters().precision_hs() == "fp32") {
-                Band(ctx_).solve<float, float>(kset_, H0, true, itsol_tol);
+                Band(ctx_).solve<float, float>(kset_, H0, iter_solver_tol__);
             } else {
-                Band(ctx_).solve<float, double>(kset_, H0, true, itsol_tol);
+                Band(ctx_).solve<float, double>(kset_, H0, iter_solver_tol__);
             }
             /* find band occupancies */
             kset_.find_band_occupancies<float>();
@@ -220,9 +230,9 @@ json DFT_ground_state::find(double density_tol, double energy_tol, double itsol_
             RTE_THROW("not compiled with FP32 support");
 #endif
         } else {
-            Hamiltonian0<double> H0(potential_);
+            Hamiltonian0<double> H0(potential_, true);
             /* find new wave-functions */
-            Band(ctx_).solve<double, double>(kset_, H0, true, itsol_tol);
+            Band(ctx_).solve<double, double>(kset_, H0, iter_solver_tol__);
             /* find band occupancies */
             kset_.find_band_occupancies<double>();
             /* generate new density from the occupied wave-functions */
@@ -244,35 +254,37 @@ json DFT_ground_state::find(double density_tol, double energy_tol, double itsol_
             tol = eha_res / std::max(1.0, unit_cell_.num_electrons());
         }
         tol = std::min(ctx_.cfg().settings().itsol_tol_scale()[0] * tol,
-                       ctx_.cfg().settings().itsol_tol_scale()[1] * itsol_tol);
+                       ctx_.cfg().settings().itsol_tol_scale()[1] * iter_solver_tol__);
         /* tolerance can't be too small */
-        itsol_tol = std::max(ctx_.cfg().settings().itsol_tol_min(), tol);
+        iter_solver_tol__ = std::max(ctx_.cfg().settings().itsol_tol_min(), tol);
 
 #if defined(USE_FP32)
-        /* if the final precision is not equal to the current precision */
-        if (ctx_.cfg().parameters().precision_gs() == "fp64" && ctx_.cfg().parameters().precision_wf() == "fp32") {
-            /* if we reached the mimimum tolerance for fp32 */
-            if ((ctx_.cfg().settings().fp32_to_fp64_rms() == 0 && itsol_tol <= ctx_.cfg().settings().itsol_tol_min()) ||
-                (rms < ctx_.cfg().settings().fp32_to_fp64_rms())) {
-                std::cout << "switching to FP64" << std::endl;
-                ctx_.cfg().unlock();
-                ctx_.cfg().settings().itsol_tol_min(std::numeric_limits<double>::epsilon() * 10);
-                ctx_.cfg().parameters().precision_wf("fp64");
-                ctx_.cfg().parameters().precision_hs("fp64");
-                ctx_.cfg().lock();
+        if (ctx_.cfg().parameters().precision_gs() != "auto") {
+            /* if the final precision is not equal to the current precision */
+            if (ctx_.cfg().parameters().precision_gs() == "fp64" && ctx_.cfg().parameters().precision_wf() == "fp32") {
+                /* if we reached the mimimum tolerance for fp32 */
+                if ((ctx_.cfg().settings().fp32_to_fp64_rms() == 0 && itsol_tol <= ctx_.cfg().settings().itsol_tol_min()) ||
+                    (rms < ctx_.cfg().settings().fp32_to_fp64_rms())) {
+                    std::cout << "switching to FP64" << std::endl;
+                    ctx_.cfg().unlock();
+                    ctx_.cfg().settings().itsol_tol_min(std::numeric_limits<double>::epsilon() * 10);
+                    ctx_.cfg().parameters().precision_wf("fp64");
+                    ctx_.cfg().parameters().precision_hs("fp64");
+                    ctx_.cfg().lock();
 
-                for (int ikloc = 0; ikloc < kset_.spl_num_kpoints().local_size(); ikloc++) {
-                    int ik = kset_.spl_num_kpoints(ikloc);
-                    for (int ispn = 0; ispn < ctx_.num_spins(); ispn++) {
-                        kset_.get<double>(ik)->spinor_wave_functions().copy_from(device_t::CPU, ctx_.num_bands(),
-                                kset_.get<float>(ik)->spinor_wave_functions(), ispn, 0, ispn, 0);
+                    for (int ikloc = 0; ikloc < kset_.spl_num_kpoints().local_size(); ikloc++) {
+                        int ik = kset_.spl_num_kpoints(ikloc);
+                        for (int ispn = 0; ispn < ctx_.num_spins(); ispn++) {
+                            kset_.get<double>(ik)->spinor_wave_functions().copy_from(device_t::CPU, ctx_.num_bands(),
+                                                                                     kset_.get<float>(ik)->spinor_wave_functions(), ispn, 0, ispn, 0);
+                        }
                     }
-                }
-                for (int ik = 0; ik < kset_.num_kpoints(); ik++) {
-                    for (int ispn = 0; ispn < ctx_.num_spinors(); ispn++) {
-                        for (int j = 0; j < ctx_.num_bands(); j++) {
-                            kset_.get<double>(ik)->band_energy(j, ispn, kset_.get<float>(ik)->band_energy(j, ispn));
-                            kset_.get<double>(ik)->band_occupancy(j, ispn, kset_.get<float>(ik)->band_occupancy(j, ispn));
+                    for (int ik = 0; ik < kset_.num_kpoints(); ik++) {
+                        for (int ispn = 0; ispn < ctx_.num_spinors(); ispn++) {
+                            for (int j = 0; j < ctx_.num_bands(); j++) {
+                                kset_.get<double>(ik)->band_energy(j, ispn, kset_.get<float>(ik)->band_energy(j, ispn));
+                                kset_.get<double>(ik)->band_occupancy(j, ispn, kset_.get<float>(ik)->band_occupancy(j, ispn));
+                            }
                         }
                     }
                 }
@@ -293,7 +305,7 @@ json DFT_ground_state::find(double density_tol, double energy_tol, double itsol_
         }
 
         if (ctx_.cfg().parameters().use_scf_correction()) {
-            double e2 = energy_potential(rho1, potential_);
+            double e2         = energy_potential(rho1, potential_);
             this->scf_energy_ = e2 - e1;
         }
 
@@ -315,11 +327,11 @@ json DFT_ground_state::find(double density_tol, double energy_tol, double itsol_
         ctx_.message(2, __func__, out);
         /* check if the calculation has converged */
         bool converged{true};
-        converged = converged && (std::abs(eold - etot) < energy_tol);
+        converged = converged && (std::abs(eold - etot) < energy_tol__);
         if (ctx_.cfg().mixer().use_hartree()) {
-            converged = converged && (eha_res < density_tol);
+            converged = converged && (eha_res < density_tol__);
         } else {
-            converged = converged && (rms < density_tol);
+            converged = converged && (rms < density_tol__);
         }
         if (converged) {
             std::stringstream out;
@@ -338,7 +350,7 @@ json DFT_ground_state::find(double density_tol, double energy_tol, double itsol_
     print_info(out);
     ctx_.message(1, __func__, out);
 
-    if (write_state) {
+    if (write_state__) {
         ctx_.create_storage_file();
         if (ctx_.full_potential()) { // TODO: why this is necessary?
             density_.rho().fft_transform(-1);
@@ -348,13 +360,13 @@ json DFT_ground_state::find(double density_tol, double energy_tol, double itsol_
         }
         potential_.save();
         density_.save();
-        //kset_.save(storage_file_name);
+        // kset_.save(storage_file_name);
     }
 
     auto tstop = std::chrono::high_resolution_clock::now();
 
-    json dict = serialize();
-    dict["scf_time"] = std::chrono::duration_cast<std::chrono::duration<double>>(tstop - tstart).count();
+    json dict            = serialize();
+    dict["scf_time"]     = std::chrono::duration_cast<std::chrono::duration<double>>(tstop - tstart).count();
     dict["etot_history"] = etot_hist;
     if (num_iter >= 0) {
         dict["converged"]          = true;
@@ -364,7 +376,7 @@ json DFT_ground_state::find(double density_tol, double energy_tol, double itsol_
         dict["converged"] = false;
     }
 
-    //if (ctx_.control().verification_ >= 1) {
+    // if (ctx_.control().verification_ >= 1) {
     //    check_scf_density();
     //}
 
@@ -376,122 +388,47 @@ json DFT_ground_state::find(double density_tol, double energy_tol, double itsol_
     return dict;
 }
 
-void DFT_ground_state::print_info(std::ostream& out__) const
+void
+DFT_ground_state::print_info(std::ostream& out__) const
 {
-    double evalsum1 = kset_.valence_eval_sum();
-    double evalsum2 = core_eval_sum(ctx_.unit_cell());
-    double s_sum    = kset_.entropy_sum();
-    double ekin     = energy_kin(ctx_, kset_, density_, potential_);
-    double evxc     = energy_vxc(density_, potential_);
-    double eexc     = energy_exc(density_, potential_);
-    double ebxc     = energy_bxc(density_, potential_);
-    double evha     = energy_vha(potential_);
-    double etot     = total_energy();
-    double gap      = kset_.band_gap() * ha2ev;
-    double ef       = kset_.energy_fermi();
-    double enuc     = energy_enuc(ctx_, potential_);
+    double evalsum1     = kset_.valence_eval_sum();
+    double evalsum2     = core_eval_sum(ctx_.unit_cell());
+    double s_sum        = kset_.entropy_sum();
+    double ekin         = energy_kin(ctx_, kset_, density_, potential_);
+    double evxc         = energy_vxc(density_, potential_);
+    double eexc         = energy_exc(density_, potential_);
+    double ebxc         = energy_bxc(density_, potential_);
+    double evha         = energy_vha(potential_);
+    double hub_one_elec = one_electron_energy_hubbard(density_, potential_);
+    double etot         = total_energy();
+    double gap          = kset_.band_gap() * ha2ev;
+    double ef           = kset_.energy_fermi();
+    double enuc         = energy_enuc(ctx_, potential_);
 
-    double one_elec_en = evalsum1 - (evxc + evha);
+    double one_elec_en = evalsum1 - (evxc + evha + ebxc);
 
     if (ctx_.electronic_structure_method() == electronic_structure_method_t::pseudopotential) {
         one_elec_en -= potential_.PAW_one_elec_energy(density_);
+        one_elec_en -= hub_one_elec;
     }
 
-    auto result = density_.rho().integrate();
+    auto draw_bar = [&](int w) { out__ << std::setfill('-') << std::setw(w) << '-' << std::setfill(' ') << std::endl; };
 
-    auto total_charge = std::get<0>(result);
-    auto it_charge    = std::get<1>(result);
-    auto mt_charge    = std::get<2>(result);
-
-    auto result_mag = density_.get_magnetisation();
-    auto total_mag  = std::get<0>(result_mag);
-    auto it_mag     = std::get<1>(result_mag);
-    auto mt_mag     = std::get<2>(result_mag);
-
-    auto draw_bar = [&](int w)
-    {
-        out__ << std::setfill('-') << std::setw(w) << '-' << std::setfill(' ') << std::endl;
-    };
-
-    auto write_vector = [&](vector3d<double> v__)
-    {
-        out__ << "["
-              << std::setw(9) << std::setprecision(5) << std::fixed << v__[0] << ", "
-              << std::setw(9) << std::setprecision(5) << std::fixed << v__[1] << ", "
-              << std::setw(9) << std::setprecision(5) << std::fixed << v__[2] << "]";
-    };
-
-    out__ << "Charges and magnetic moments" << std::endl;
-    draw_bar(80);
-    if (ctx_.full_potential()) {
-        double total_core_leakage{0.0};
-        out__ << "atom      charge    core leakage";
-        if (ctx_.num_mag_dims()) {
-            out__ << "                 moment                |moment|";
-        }
-        out__ << std::endl;
-        draw_bar(80);
-
-        for (int ia = 0; ia < unit_cell_.num_atoms(); ia++) {
-            double core_leakage = unit_cell_.atom(ia).symmetry_class().core_leakage();
-            total_core_leakage += core_leakage;
-            out__ << std::setw(4) << ia
-                  << std::setw(12) << std::setprecision(6) << std::fixed << mt_charge[ia]
-                  << std::setw(16) << std::setprecision(6) << std::scientific << core_leakage;
-            if (ctx_.num_mag_dims()) {
-                vector3d<double> v(mt_mag[ia]);
-                out__ << "  ";
-                write_vector(v);
-                out__ << std::setw(12) << std::setprecision(6) << std::fixed << v.length();
-            }
-            out__ << std::endl;
-        }
-        out__ << std::endl;
-        out__ << "total core leakage    : " << std::setprecision(8) << std::scientific << total_core_leakage << std::endl
-              << "interstitial charge   : " << std::setprecision(6) << std::fixed << it_charge << std::endl;
-        if (ctx_.num_mag_dims()) {
-            vector3d<double> v(it_mag);
-            out__ << "interstitial moment   : ";
-            write_vector(v);
-            out__ <<  ", magnitude : " << std::setprecision(6) << std::fixed << v.length() << std::endl;
-        }
-    } else {
-        if (ctx_.num_mag_dims()) {
-            out__ << "atom                moment                |moment|" << std::endl;
-            draw_bar(80);
-
-            for (int ia = 0; ia < unit_cell_.num_atoms(); ia++) {
-                vector3d<double> v(mt_mag[ia]);
-                out__ << std::setw(4) << ia << " ";
-                write_vector(v);
-                out__ << std::setw(12) << std::setprecision(6) << std::fixed << v.length() << std::endl;
-            }
-            out__ << std::endl;
-        }
-    }
-    out__ << "total charge          : " << std::setprecision(6) << std::fixed << total_charge << std::endl;
-
-    if (ctx_.num_mag_dims()) {
-        vector3d<double> v(total_mag);
-        out__ << "total moment          : ";
-        write_vector(v);
-        out__ << ", magnitude : " << std::setprecision(6) << std::fixed << v.length() << std::endl;
-    }
+    density_.print_info(out__);
 
     out__ << std::endl;
     out__ << "Energy" << std::endl;
     draw_bar(80);
 
-    auto write_energy = [&](std::string label__, double value__)
-    {
-        out__ << std::left << std::setw(30) << label__ << " : "
-              << std::right << std::setw(16) << std::setprecision(8) << std::fixed << value__ << std::endl;
+    auto write_energy = [&](std::string label__, double value__) {
+        out__ << std::left << std::setw(30) << label__ << " : " << std::right << std::setw(16) << std::setprecision(8)
+              << std::fixed << value__ << std::endl;
     };
 
     auto write_energy2 = [&](std::string label__, double value__)
     {
         out__ << std::left << std::setw(30) << label__ << " : " << std::right
-              << std::setw(16) << std::setprecision(8) << std::fixed << value__ << " (Ha), " 
+              << std::setw(16) << std::setprecision(8) << std::fixed << value__ << " (Ha), "
               << std::setw(16) << std::setprecision(8) << std::fixed << value__ * 2 << " (Ry)" << std::endl;
     };
 
@@ -517,6 +454,7 @@ void DFT_ground_state::print_info(std::ostream& out__) const
     if (ctx_.hubbard_correction()) {
         auto e = ::sirius::energy(density_.occupation_matrix());
         write_energy2("Hubbard energy", e);
+        write_energy2("Hubbard one-el contribution", hub_one_elec);
     }
     write_energy2("Total energy", etot);
     out__ << std::endl;
