@@ -44,7 +44,8 @@ Force::Force(Simulation_context& ctx__, Density& density__, Potential& potential
 {
 }
 
-void Force::symmetrize(mdarray<double, 2>& forces__) const
+void
+Force::symmetrize(mdarray<double, 2>& forces__) const
 {
     if (!ctx_.use_symmetry()) {
         return;
@@ -57,16 +58,16 @@ void Force::symmetrize(mdarray<double, 2>& forces__) const
 
     for (int isym = 0; isym < ctx_.unit_cell().symmetry().size(); isym++) {
         auto Rc = dot(dot(ctx_.unit_cell().symmetry().lattice_vectors(),
-                  matrix3d<double>(ctx_.unit_cell().symmetry()[isym].spg_op.R)),
-                  ctx_.unit_cell().symmetry().inverse_lattice_vectors());
+                          matrix3d<double>(ctx_.unit_cell().symmetry()[isym].spg_op.R)),
+                      ctx_.unit_cell().symmetry().inverse_lattice_vectors());
 
         for (int ia = 0; ia < ctx_.unit_cell().num_atoms(); ia++) {
             vector3d<double> force_ia(&forces__(0, ia));
-            int ja = ctx_.unit_cell().symmetry()[isym].spg_op.sym_atom[ia];
+            int ja        = ctx_.unit_cell().symmetry()[isym].spg_op.sym_atom[ia];
             auto location = ctx_.unit_cell().spl_num_atoms().location(ja);
             if (location.rank == ctx_.comm().rank()) {
                 auto force_ja = dot(Rc, force_ia);
-                for (int x: {0, 1, 2}) {
+                for (int x : {0, 1, 2}) {
                     sym_forces(x, location.local_index) += force_ja[x];
                 }
             }
@@ -122,14 +123,15 @@ sddk::mdarray<double, 2> const& Force::calc_forces_nonloc()
 }
 
 template <typename T>
-void Force::add_k_point_contribution(K_point<real_type<T>>& kp__, mdarray<double, 2>& forces__) const
+void
+Force::add_k_point_contribution(K_point<real_type<T>>& kp__, sddk::mdarray<double, 2>& forces__) const
 {
     /* if there are no beta projectors then get out there */
     if (ctx_.unit_cell().mt_lo_basis_size() == 0) {
         return;
     }
 
-    Beta_projectors_gradient<real_type<T>> bp_grad(ctx_, kp__.gkvec(), kp__.igk_loc(), kp__.beta_projectors());
+    Beta_projectors_gradient<real_type<T>> bp_grad(ctx_, kp__.gkvec(), kp__.beta_projectors());
     if (is_device_memory(ctx_.preferred_memory_t())) {
         int nbnd = ctx_.num_bands();
         for (int ispn = 0; ispn < ctx_.num_spins(); ispn++) {
@@ -146,7 +148,7 @@ void Force::add_k_point_contribution(K_point<real_type<T>>& kp__, mdarray<double
 
     nlf.add_k_point_contribution(kp__, f);
     for (int ia = 0; ia < ctx_.unit_cell().num_atoms(); ia++) {
-        for (int x: {0, 1, 2}) {
+        for (int x : {0, 1, 2}) {
             forces__(x, ia) += f(x, ia);
         }
     }
@@ -158,7 +160,8 @@ void Force::add_k_point_contribution(K_point<real_type<T>>& kp__, mdarray<double
     }
 }
 
-void Force::compute_dmat(K_point<double>* kp__, dmatrix<double_complex>& dm__) const
+void
+Force::compute_dmat(K_point<double>* kp__, dmatrix<double_complex>& dm__) const
 {
     dm__.zero();
 
@@ -182,9 +185,10 @@ void Force::compute_dmat(K_point<double>* kp__, dmatrix<double_complex>& dm__) c
                     }
                 }
 
-                linalg(linalg_t::scalapack).gemm('N', 'T', ctx_.num_fv_states(), ctx_.num_fv_states(), ctx_.num_bands(),
-                                                  &linalg_const<double_complex>::one(), ev1, 0, 0, ev, 0, 0,
-                                                  &linalg_const<double_complex>::one(), dm__, 0, 0);
+                linalg(linalg_t::scalapack)
+                    .gemm('N', 'T', ctx_.num_fv_states(), ctx_.num_fv_states(), ctx_.num_bands(),
+                          &linalg_const<double_complex>::one(), ev1, 0, 0, ev, 0, 0,
+                          &linalg_const<double_complex>::one(), dm__, 0, 0);
             }
         } else {
             dmatrix<double_complex> ev1(ctx_.num_bands(), ctx_.num_bands(), ctx_.blacs_grid(), ctx_.cyclic_block_size(),
@@ -201,15 +205,17 @@ void Force::compute_dmat(K_point<double>* kp__, dmatrix<double_complex>& dm__) c
             for (int ispn = 0; ispn < ctx_.num_spins(); ispn++) {
                 int offs = ispn * ctx_.num_fv_states();
 
-                linalg(linalg_t::scalapack).gemm('N', 'T', ctx_.num_fv_states(), ctx_.num_fv_states(), ctx_.num_bands(),
-                                                  &linalg_const<double_complex>::one(), ev1, offs, 0, ev, offs, 0,
-                                                  &linalg_const<double_complex>::one(), dm__, 0, 0);
+                linalg(linalg_t::scalapack)
+                    .gemm('N', 'T', ctx_.num_fv_states(), ctx_.num_fv_states(), ctx_.num_bands(),
+                          &linalg_const<double_complex>::one(), ev1, offs, 0, ev, offs, 0,
+                          &linalg_const<double_complex>::one(), dm__, 0, 0);
             }
         }
     }
 }
 
-mdarray<double, 2> const& Force::calc_forces_total()
+mdarray<double, 2> const&
+Force::calc_forces_total()
 {
     forces_total_ = mdarray<double, 2>(3, ctx_.unit_cell().num_atoms());
     if (ctx_.full_potential()) {
@@ -242,7 +248,8 @@ mdarray<double, 2> const& Force::calc_forces_total()
     return forces_total_;
 }
 
-mdarray<double, 2> const& Force::calc_forces_ibs()
+mdarray<double, 2> const&
+Force::calc_forces_ibs()
 {
     forces_ibs_ = mdarray<double, 2>(3, ctx_.unit_cell().num_atoms());
     forces_ibs_.zero();
@@ -256,9 +263,9 @@ mdarray<double, 2> const& Force::calc_forces_ibs()
         }
     }
 
-    Hamiltonian0<double> H0(potential_);
+    Hamiltonian0<double> H0(potential_, true);
     for (int ikloc = 0; ikloc < kset_.spl_num_kpoints().local_size(); ikloc++) {
-        int ik = kset_.spl_num_kpoints(ikloc);
+        int ik  = kset_.spl_num_kpoints(ikloc);
         auto hk = H0(*kset_.get<double>(ik));
         add_ibs_force(kset_.get<double>(ik), hk, ffac, forces_ibs_);
     }
@@ -269,14 +276,15 @@ mdarray<double, 2> const& Force::calc_forces_ibs()
         std::printf("ibs force\n");
         for (int ia = 0; ia < ctx_.unit_cell().num_atoms(); ia++) {
             std::printf("ia : %i, IBS : %12.6f %12.6f %12.6f\n", ia, forces_ibs_(0, ia), forces_ibs_(1, ia),
-                   forces_ibs_(2, ia));
+                        forces_ibs_(2, ia));
         }
     }
 
     return forces_ibs_;
 }
 
-mdarray<double, 2> const& Force::calc_forces_rho()
+mdarray<double, 2> const&
+Force::calc_forces_rho()
 {
 
     forces_rho_ = mdarray<double, 2>(3, ctx_.unit_cell().num_atoms());
@@ -294,14 +302,15 @@ mdarray<double, 2> const& Force::calc_forces_rho()
     if (ctx_.verbosity() > 2 && ctx_.comm().rank() == 0) {
         std::printf("rho force\n");
         for (int ia = 0; ia < ctx_.unit_cell().num_atoms(); ia++) {
-            std::printf("ia : %i, density contribution : %12.6f %12.6f %12.6f\n", ia, forces_rho_(0, ia), forces_rho_(1, ia),
-                   forces_rho_(2, ia));
+            std::printf("ia : %i, density contribution : %12.6f %12.6f %12.6f\n", ia, forces_rho_(0, ia),
+                        forces_rho_(1, ia), forces_rho_(2, ia));
         }
     }
     return forces_rho_;
 }
 
-mdarray<double, 2> const& Force::calc_forces_hf()
+mdarray<double, 2> const&
+Force::calc_forces_hf()
 {
     forces_hf_ = mdarray<double, 2>(3, ctx_.unit_cell().num_atoms());
     forces_hf_.zero();
@@ -319,15 +328,16 @@ mdarray<double, 2> const& Force::calc_forces_hf()
     if (ctx_.cfg().control().verbosity() > 2 && ctx_.comm().rank() == 0) {
         std::printf("H-F force\n");
         for (int ia = 0; ia < ctx_.unit_cell().num_atoms(); ia++) {
-            std::printf("ia : %i, Hellmann–Feynman : %12.6f %12.6f %12.6f\n", ia, forces_hf_(0, ia), forces_hf_(1, ia),
-                   forces_hf_(2, ia));
+            std::printf("ia : %i, Hellmann Feynman : %12.6f %12.6f %12.6f\n", ia, forces_hf_(0, ia), forces_hf_(1, ia),
+                        forces_hf_(2, ia));
         }
     }
 
     return forces_hf_;
 }
 
-mdarray<double, 2> const& Force::calc_forces_hubbard()
+mdarray<double, 2> const&
+Force::calc_forces_hubbard()
 {
     PROFILE("sirius::Force::hubbard_force");
     forces_hubbard_ = mdarray<double, 2>(3, ctx_.unit_cell().num_atoms());
@@ -355,10 +365,12 @@ mdarray<double, 2> const& Force::calc_forces_hubbard()
         kset_.comm().allreduce(forces_hubbard_.at(memory_t::host), 3 * ctx_.unit_cell().num_atoms());
     }
 
+    symmetrize(forces_hubbard_);
     return forces_hubbard_;
 }
 
-mdarray<double, 2> const& Force::calc_forces_ewald()
+mdarray<double, 2> const&
+Force::calc_forces_ewald()
 {
     PROFILE("sirius::Force::calc_forces_ewald");
 
@@ -393,7 +405,7 @@ mdarray<double, 2> const& Force::calc_forces_ewald()
         for (int igloc = ig0; igloc < ctx_.gvec().count(); igloc++) {
             int ig = ctx_.gvec().offset() + igloc;
 
-            double g2 = std::pow(ctx_.gvec().gvec_len(ig), 2);
+            double g2 = std::pow(ctx_.gvec().gvec_len<index_domain_t::local>(igloc), 2);
 
             /* cartesian form for getting cartesian force components */
             auto gvec_cart = ctx_.gvec().gvec_cart<index_domain_t::local>(igloc);
@@ -436,7 +448,8 @@ mdarray<double, 2> const& Force::calc_forces_ewald()
     return forces_ewald_;
 }
 
-mdarray<double, 2> const& Force::calc_forces_us()
+mdarray<double, 2> const&
+Force::calc_forces_us()
 {
     PROFILE("sirius::Force::calc_forces_us");
 
@@ -506,17 +519,16 @@ mdarray<double, 2> const& Force::calc_forces_us()
 
                 /* multiply tmp matrices, or sum over G */
                 linalg(la).gemm('N', 'T', nbf * (nbf + 1) / 2, atom_type.num_atoms(), 2 * ctx_.gvec().count(),
-                    &linalg_const<double>::one(),
-                    aug_op->q_pw().at(memory_t::host), aug_op->q_pw().ld(),
-                    v_tmp.at(memory_t::host), v_tmp.ld(),
-                    &linalg_const<double>::zero(),
-                    tmp.at(memory_t::host), tmp.ld());
+                                &linalg_const<double>::one(), aug_op->q_pw().at(memory_t::host), aug_op->q_pw().ld(),
+                                v_tmp.at(memory_t::host), v_tmp.ld(), &linalg_const<double>::zero(),
+                                tmp.at(memory_t::host), tmp.ld());
 
                 #pragma omp parallel for
                 for (int ia = 0; ia < atom_type.num_atoms(); ia++) {
                     for (int i = 0; i < nbf * (nbf + 1) / 2; i++) {
                         forces_us_(ivec, atom_type.atom_id(ia)) += ctx_.unit_cell().omega() * reduce_g_fact *
-                                                                   dm(i, ia, ispin) * aug_op->sym_weight(i) * tmp(i, ia);
+                                                                   dm(i, ia, ispin) * aug_op->sym_weight(i) *
+                                                                   tmp(i, ia);
                     }
                 }
             }
@@ -530,7 +542,8 @@ mdarray<double, 2> const& Force::calc_forces_us()
     return forces_us_;
 }
 
-mdarray<double, 2> const& Force::calc_forces_scf_corr()
+mdarray<double, 2> const&
+Force::calc_forces_scf_corr()
 {
     PROFILE("sirius::Force::calc_forces_scf_corr");
 
@@ -562,15 +575,15 @@ mdarray<double, 2> const& Force::calc_forces_scf_corr()
         int iat = atom.type_id();
 
         for (int igloc = ig0; igloc < gvec_count; igloc++) {
-            int ig = gvec_offset + igloc;
+            int ig   = gvec_offset + igloc;
             int igsh = ctx_.gvec().shell(ig);
 
             /* cartesian form for getting cartesian force components */
             auto gvec_cart = gvec.gvec_cart<index_domain_t::local>(igloc);
 
             /* scalar part of a force without multipying by G-vector */
-            double_complex z = fact * fourpi * ff(igsh, iat) *
-                               std::conj(dveff.f_pw_local(igloc) * ctx_.gvec_phase_factor(ig, ia));
+            double_complex z =
+                fact * fourpi * ff(igsh, iat) * std::conj(dveff.f_pw_local(igloc) * ctx_.gvec_phase_factor(ig, ia));
 
             /* get force components multiplying by cartesian G-vector */
             for (int x : {0, 1, 2}) {
@@ -585,7 +598,8 @@ mdarray<double, 2> const& Force::calc_forces_scf_corr()
     return forces_scf_corr_;
 }
 
-mdarray<double, 2> const& Force::calc_forces_core()
+mdarray<double, 2> const&
+Force::calc_forces_core()
 {
     PROFILE("sirius::Force::calc_forces_core");
 
@@ -621,15 +635,15 @@ mdarray<double, 2> const& Force::calc_forces_core()
         int iat = atom.type_id();
 
         for (int igloc = ctx_.gvec().skip_g0(); igloc < gvec_count; igloc++) {
-            int ig = gvec_offset + igloc;
+            int ig    = gvec_offset + igloc;
             auto igsh = ctx_.gvec().shell(ig);
 
             /* cartesian form for getting cartesian force components */
             auto gvec_cart = gvecs.gvec_cart<index_domain_t::local>(igloc);
 
             /* scalar part of a force without multipying by G-vector */
-            double_complex z = fact * fourpi * ff(igsh, iat) *
-                               std::conj(xc_pot.f_pw_local(igloc) * ctx_.gvec_phase_factor(ig, ia));
+            double_complex z =
+                fact * fourpi * ff(igsh, iat) * std::conj(xc_pot.f_pw_local(igloc) * ctx_.gvec_phase_factor(ig, ia));
 
             /* get force components multiplying by cartesian G-vector */
             for (int x : {0, 1, 2}) {
@@ -644,10 +658,13 @@ mdarray<double, 2> const& Force::calc_forces_core()
     return forces_core_;
 }
 
-void Force::hubbard_force_add_k_contribution_collinear(K_point<double>& kp__, Q_operator<double>& q_op__, mdarray<double, 2>& forceh_)
+void
+Force::hubbard_force_add_k_contribution_collinear(K_point<double>& kp__, Q_operator<double>& q_op__,
+                                                  mdarray<double, 2>& forceh__)
 {
-    mdarray<double_complex, 6> dn(potential_.U().max_number_of_orbitals_per_atom(), potential_.U().max_number_of_orbitals_per_atom(), 2,
-                                  ctx_.unit_cell().num_atoms(), 3, ctx_.unit_cell().num_atoms());
+    auto r = ctx_.unit_cell().num_hubbard_wf();
+
+    sddk::mdarray<double_complex, 5> dn(r.first, r.first, ctx_.num_spins(), 3, ctx_.unit_cell().num_atoms());
 
     potential_.U().compute_occupancies_derivatives(kp__, q_op__, dn);
 
@@ -656,26 +673,57 @@ void Force::hubbard_force_add_k_contribution_collinear(K_point<double>& kp__, Q_
         /* compute the derivative of the occupancies numbers */
         for (int dir = 0; dir < 3; dir++) {
             double_complex d{0.0};
-            for (int ia1 = 0; ia1 < ctx_.unit_cell().num_atoms(); ia1++) {
-                auto const& atom = ctx_.unit_cell().atom(ia1);
-                if (atom.type().hubbard_correction()) {
-                    int const lmax_at = 2 * atom.type().lo_descriptor_hub(0).l + 1;
+            for (int at_lvl = 0; at_lvl < static_cast<int>(potential_.hubbard_potential().local().size()); at_lvl++) {
+                const int ia1    = potential_.hubbard_potential().atomic_orbitals(at_lvl).first;
+                const auto& atom = ctx_.unit_cell().atom(ia1);
+                const int lo     = potential_.hubbard_potential().atomic_orbitals(at_lvl).second;
+                if (atom.type().lo_descriptor_hub(lo).use_for_calculation()) {
+                    int const lmax_at = 2 * atom.type().lo_descriptor_hub(lo).l() + 1;
+                    const int offset  = potential_.hubbard_potential().offset(at_lvl);
                     for (int ispn = 0; ispn < ctx_.num_spins(); ispn++) {
-                        for (int m1 = 0; m1 < lmax_at; m1++) {
-                            for (int m2 = 0; m2 < lmax_at; m2++) {
-                                d += potential_.hubbard_potential().local(ia1)(m2, m1, ispn) *
-                                     dn(m1, m2, ispn, ia1, dir, ia);
+                        for (int m2 = 0; m2 < lmax_at; m2++) {
+                            for (int m1 = 0; m1 < lmax_at; m1++) {
+                                d += potential_.hubbard_potential().local(at_lvl)(m1, m2, ispn) *
+                                    dn(offset + m2, offset + m1, ispn, dir, ia);
                             }
                         }
                     }
                 }
             }
-            forceh_(dir, ia) -= std::real(d);
+
+            for (int i = 0; i < ctx_.cfg().hubbard().nonlocal().size(); i++) {
+                auto nl = ctx_.cfg().hubbard().nonlocal(i);
+                int ia1 = nl.atom_pair()[0];
+                int ja  = nl.atom_pair()[1];
+                // consider the links that involve atom i
+                int il  = nl.l()[0];
+                int jl  = nl.l()[1];
+                int in  = nl.n()[0];
+                int jn  = nl.n()[1];
+                auto Tr = nl.T();
+
+                auto z1           = std::exp(double_complex(0, -twopi * dot(vector3d<int>(Tr), kp__.vk())));
+                const int at_lvl1 = potential_.hubbard_potential().find_orbital_index(ia1, in, il);
+                const int at_lvl2 = potential_.hubbard_potential().find_orbital_index(ja, jn, jl);
+                const int offset1 = potential_.hubbard_potential().offset(at_lvl1);
+                const int offset2 = potential_.hubbard_potential().offset(at_lvl2);
+                for (int is = 0; is < ctx_.num_spins(); is++) {
+                    for (int m2 = 0; m2 < 2 * jl + 1; m2++) {
+                        for (int m1 = 0; m1 < 2 * il + 1; m1++) {
+                            auto result1_ = z1 * std::conj(dn(offset2 + m2, offset1 + m1, is, dir, ia)) *
+                                potential_.hubbard_potential().nonlocal(i)(m1, m2, is);
+                            d += std::real(result1_);
+                        }
+                    }
+                }
+            }
+            forceh__(dir, ia) -= std::real(d);
         }
     }
 }
 
-mdarray<double, 2> const& Force::calc_forces_vloc()
+mdarray<double, 2> const&
+Force::calc_forces_vloc()
 {
     PROFILE("sirius::Force::calc_forces_vloc");
 
@@ -705,15 +753,15 @@ mdarray<double, 2> const& Force::calc_forces_vloc()
         int iat = atom.type_id();
 
         for (int igloc = 0; igloc < gvec_count; igloc++) {
-            int ig = gvec_offset + igloc;
+            int ig   = gvec_offset + igloc;
             int igsh = ctx_.gvec().shell(ig);
 
             /* cartesian form for getting cartesian force components */
             auto gvec_cart = gvecs.gvec_cart<index_domain_t::local>(igloc);
 
             /* scalar part of a force without multiplying by G-vector */
-            double_complex z = fact * fourpi * ff(igsh, iat) *
-                               std::conj(valence_rho.f_pw_local(igloc)) * std::conj(ctx_.gvec_phase_factor(ig, ia));
+            double_complex z = fact * fourpi * ff(igsh, iat) * std::conj(valence_rho.f_pw_local(igloc)) *
+                               std::conj(ctx_.gvec_phase_factor(ig, ia));
 
             /* get force components multiplying by cartesian G-vector  */
             for (int x : {0, 1, 2}) {
@@ -729,7 +777,8 @@ mdarray<double, 2> const& Force::calc_forces_vloc()
     return forces_vloc_;
 }
 
-mdarray<double, 2> const& Force::calc_forces_usnl()
+mdarray<double, 2> const&
+Force::calc_forces_usnl()
 {
     calc_forces_us();
     calc_forces_nonloc();
@@ -744,7 +793,9 @@ mdarray<double, 2> const& Force::calc_forces_usnl()
     return forces_usnl_;
 }
 
-void Force::add_ibs_force(K_point<double>* kp__, Hamiltonian_k<double>& Hk__, mdarray<double, 2>& ffac__, mdarray<double, 2>& forcek__) const
+void
+Force::add_ibs_force(K_point<double>* kp__, Hamiltonian_k<double>& Hk__, mdarray<double, 2>& ffac__,
+                     mdarray<double, 2>& forcek__) const
 {
     PROFILE("sirius::Force::ibs_force");
 
@@ -796,23 +847,26 @@ void Force::add_ibs_force(K_point<double>* kp__, Hamiltonian_k<double>& Hk__, md
         Hk__.H0().apply_hmt_to_apw<spin_block_t::nm>(atom, kp__->num_gkvec_col(), alm_col, halm_col);
 
         /* apw-apw block of the overlap matrix */
-        linalg(linalg_t::blas).gemm('N', 'T', kp__->num_gkvec_row(), kp__->num_gkvec_col(), type.mt_aw_basis_size(),
-            &linalg_const<double_complex>::one(), alm_row.at(memory_t::host), alm_row.ld(),
-            alm_col.at(memory_t::host), alm_col.ld(), &linalg_const<double_complex>::zero(),
-            o.at(memory_t::host), o.ld());
+        linalg(linalg_t::blas)
+            .gemm('N', 'T', kp__->num_gkvec_row(), kp__->num_gkvec_col(), type.mt_aw_basis_size(),
+                  &linalg_const<double_complex>::one(), alm_row.at(memory_t::host), alm_row.ld(),
+                  alm_col.at(memory_t::host), alm_col.ld(), &linalg_const<double_complex>::zero(), o.at(memory_t::host),
+                  o.ld());
 
         /* apw-apw block of the Hamiltonian matrix */
-        linalg(linalg_t::blas).gemm('N', 'T', kp__->num_gkvec_row(), kp__->num_gkvec_col(), type.mt_aw_basis_size(),
-            &linalg_const<double_complex>::one(), alm_row.at(memory_t::host), alm_row.ld(), halm_col.at(memory_t::host),
-            halm_col.ld(), &linalg_const<double_complex>::zero(), h.at(memory_t::host), h.ld());
+        linalg(linalg_t::blas)
+            .gemm('N', 'T', kp__->num_gkvec_row(), kp__->num_gkvec_col(), type.mt_aw_basis_size(),
+                  &linalg_const<double_complex>::one(), alm_row.at(memory_t::host), alm_row.ld(),
+                  halm_col.at(memory_t::host), halm_col.ld(), &linalg_const<double_complex>::zero(),
+                  h.at(memory_t::host), h.ld());
 
         int iat = type.id();
 
         for (int igk_col = 0; igk_col < kp__->num_gkvec_col(); igk_col++) { // loop over columns
-            auto gvec_col       = kp__->gkvec().gvec(kp__->igk_col(igk_col));
+            auto gvec_col       = kp__->gkvec().gvec<index_domain_t::global>(kp__->igk_col(igk_col));
             auto gkvec_col_cart = kp__->gkvec().gkvec_cart<index_domain_t::global>(kp__->igk_col(igk_col));
             for (int igk_row = 0; igk_row < kp__->num_gkvec_row(); igk_row++) { // for each column loop over rows
-                auto gvec_row       = kp__->gkvec().gvec(kp__->igk_row(igk_row));
+                auto gvec_row       = kp__->gkvec().gvec<index_domain_t::global>(kp__->igk_row(igk_row));
                 auto gkvec_row_cart = kp__->gkvec().gkvec_cart<index_domain_t::global>(kp__->igk_row(igk_row));
 
                 int ig12 = ctx_.gvec().index_g12(gvec_row, gvec_col);
@@ -830,9 +884,9 @@ void Force::add_ibs_force(K_point<double>* kp__, Hamiltonian_k<double>& Hk__, md
 
         for (int x = 0; x < 3; x++) {
             for (int igk_col = 0; igk_col < kp__->num_gkvec_col(); igk_col++) { // loop over columns
-                auto gvec_col = kp__->gkvec().gvec(kp__->igk_col(igk_col));
+                auto gvec_col = kp__->gkvec().gvec<index_domain_t::global>(kp__->igk_col(igk_col));
                 for (int igk_row = 0; igk_row < kp__->num_gkvec_row(); igk_row++) { // loop over rows
-                    auto gvec_row = kp__->gkvec().gvec(kp__->igk_row(igk_row));
+                    auto gvec_row = kp__->gkvec().gvec<index_domain_t::global>(kp__->igk_row(igk_row));
                     /* compute index of G-G' */
                     int ig12 = ctx_.gvec().index_g12(gvec_row, gvec_col);
                     /* get G-G' */
@@ -869,8 +923,9 @@ void Force::add_ibs_force(K_point<double>* kp__, Hamiltonian_k<double>& Hk__, md
             }
 
             /* zm1 = dO * V */
-            linalg(linalg_t::scalapack).gemm('N', 'N', ngklo, nfv, ngklo, &linalg_const<double_complex>::one(),
-                o1, 0, 0, fv_evec, 0, 0, &linalg_const<double_complex>::zero(), zm1, 0, 0);
+            linalg(linalg_t::scalapack)
+                .gemm('N', 'N', ngklo, nfv, ngklo, &linalg_const<double_complex>::one(), o1, 0, 0, fv_evec, 0, 0,
+                      &linalg_const<double_complex>::zero(), zm1, 0, 0);
             /* multiply by energy: zm1 = E * (dO * V)  */
             for (int i = 0; i < zm1.num_cols_local(); i++) {
                 int ist = zm1.icol(i);
@@ -879,12 +934,14 @@ void Force::add_ibs_force(K_point<double>* kp__, Hamiltonian_k<double>& Hk__, md
                 }
             }
             /* compute zm1 = dH * V - E * (dO * V) */
-            linalg(linalg_t::scalapack).gemm('N', 'N', ngklo, nfv, ngklo, &linalg_const<double_complex>::one(),
-                h1, 0, 0, fv_evec, 0, 0, &linalg_const<double_complex>::m_one(), zm1, 0, 0);
+            linalg(linalg_t::scalapack)
+                .gemm('N', 'N', ngklo, nfv, ngklo, &linalg_const<double_complex>::one(), h1, 0, 0, fv_evec, 0, 0,
+                      &linalg_const<double_complex>::m_one(), zm1, 0, 0);
 
             /* compute zf = V^{+} * zm1 = V^{+} * (dH * V - E * (dO * V)) */
-            linalg(linalg_t::scalapack).gemm('C', 'N', nfv, nfv, ngklo, &linalg_const<double_complex>::one(),
-                fv_evec, 0, 0, zm1, 0, 0, &linalg_const<double_complex>::zero(), zf, 0, 0);
+            linalg(linalg_t::scalapack)
+                .gemm('C', 'N', nfv, nfv, ngklo, &linalg_const<double_complex>::one(), fv_evec, 0, 0, zm1, 0, 0,
+                      &linalg_const<double_complex>::zero(), zf, 0, 0);
 
             for (int i = 0; i < dm.num_cols_local(); i++) {
                 for (int j = 0; j < dm.num_rows_local(); j++) {
@@ -895,13 +952,14 @@ void Force::add_ibs_force(K_point<double>* kp__, Hamiltonian_k<double>& Hk__, md
     } // ia
 }
 
-void Force::print_info()
+void
+Force::print_info()
 {
     if (ctx_.comm().rank() == 0) {
         auto print_forces = [&](mdarray<double, 2> const& forces) {
             for (int ia = 0; ia < ctx_.unit_cell().num_atoms(); ia++) {
                 std::printf("atom %4i    force = %15.7f  %15.7f  %15.7f \n", ctx_.unit_cell().atom(ia).type_id(),
-                       forces(0, ia), forces(1, ia), forces(2, ia));
+                            forces(0, ia), forces(1, ia), forces(2, ia));
             }
         };
 
@@ -924,7 +982,7 @@ void Force::print_info()
         print_forces(forces_ewald());
 
         if (ctx_.hubbard_correction()) {
-            std::printf("===== Ewald forces from hubbard correction =====\n");
+            std::printf("===== contribution from Hubbard correction =====\n");
             print_forces(forces_hubbard());
         }
     }
