@@ -32,31 +32,61 @@ Hubbard_matrix::Hubbard_matrix(Simulation_context& ctx__)
 {
     if (!ctx_.full_potential() && ctx_.hubbard_correction()) {
 
-        local_ = std::vector<sddk::mdarray<double_complex, 3>>(ctx_.unit_cell().num_atoms());
-
+        /* first compute the number of atomic levels involved in the hubbard correction */
+        int num_atomic_level{0};
+        atomic_orbitals_.clear();
         for (int ia = 0; ia < ctx_.unit_cell().num_atoms(); ia++) {
-            if (ctx_.unit_cell().atom(ia).type().hubbard_correction()) {
-                int nb = ctx_.unit_cell().atom(ia).type().indexb_hub().size();
-                local_[ia] = sddk::mdarray<double_complex, 3>(nb, nb, 4);
-                local_[ia].zero();
+            auto& atom_type = ctx_.unit_cell().atom(ia).type();
+            if (atom_type.hubbard_correction()) {
+                num_atomic_level += atom_type.lo_descriptor_hub().size();
+
+                for (int lo = 0; lo < static_cast<int>(atom_type.lo_descriptor_hub().size()); lo++) {
+                    atomic_orbitals_.push_back(std::make_pair(ia, lo));
+                }
             }
         }
+
+        local_ = std::vector<sddk::mdarray<double_complex, 3>>(num_atomic_level);
+
+        /* the offsets here match the offsets of the hubbard wave functions but
+         * are more fine grained. The offsets of the hubbard wave functions are
+         * for atom while here they are for each atomic level. Since all atomic
+         * level of a given atom are next to each other, the offset of the first
+         * atomic level of a given atom has the same value than the offset
+         * giving the position of the first hubbard wave function of this
+         * atom. */
+        offset_ = std::vector<int>(num_atomic_level, 0);
+
+        int size{0};
+        for (int at_lvl = 0; at_lvl < static_cast<int>(local_.size()); at_lvl++) {
+            offset_[at_lvl] = size;
+            int ia          = atomic_orbitals_[at_lvl].first;
+            auto& atom_type = ctx_.unit_cell().atom(ia).type();
+            int lo_ind      = atomic_orbitals_[at_lvl].second;
+            int l           = atom_type.lo_descriptor_hub(lo_ind).l();
+            int mmax        = 2 * l + 1;
+
+            local_[at_lvl] = sddk::mdarray<double_complex, 3>(mmax, mmax, 4, memory_t::host, "local_hubbard");
+            local_[at_lvl].zero();
+            size += mmax;
+        }
+
+        nonlocal_.clear();
         nonlocal_ = std::vector<sddk::mdarray<double_complex, 3>>(ctx_.cfg().hubbard().nonlocal().size());
-        for (int i = 0; i < ctx_.cfg().hubbard().nonlocal().size(); i++) {
-            auto nl = ctx_.cfg().hubbard().nonlocal(i);
-            int ia = nl.atom_pair()[0];
-            int ja = nl.atom_pair()[1];
-            int il = nl.l()[0];
-            int jl = nl.l()[1];
-            int ib = ctx_.unit_cell().atom(ia).type().indexr_hub().subshell_size(il, 0);
-            int jb = ctx_.unit_cell().atom(ja).type().indexr_hub().subshell_size(jl, 0);
-            nonlocal_[i] = sddk::mdarray<double_complex, 3>(ib, jb, 4);
+        for (int i = 0; i < static_cast<int>(ctx_.cfg().hubbard().nonlocal().size()); i++) {
+            auto nl      = ctx_.cfg().hubbard().nonlocal(i);
+            int il       = nl.l()[0];
+            int jl       = nl.l()[1];
+            nonlocal_[i] = sddk::mdarray<double_complex, 3>(2 * il + 1, 2 * jl + 1, 4, memory_t::host,
+                    "nonlocal_hubbard");
             nonlocal_[i].zero();
         }
     }
 }
 
-void Hubbard_matrix::access(std::string const& what__, double_complex* occ__, int ld__) {
+void
+Hubbard_matrix::access(std::string const& what__, double_complex* occ__, int ld__)
+{
     if (!(what__ == "get" || what__ == "set")) {
         std::stringstream s;
         s << "wrong access label: " << what__;
@@ -74,19 +104,22 @@ void Hubbard_matrix::access(std::string const& what__, double_complex* occ__, in
         occ_mtrx.zero();
     }
 
-    for (int ia = 0; ia < ctx_.unit_cell().num_atoms(); ia++) {
-        auto& atom = ctx_.unit_cell().atom(ia);
-        if (atom.type().hubbard_correction()) {
-            const int l = ctx_.unit_cell().atom(ia).type().indexr_hub().am(0).l();
+    for (int at_lvl = 0; at_lvl < static_cast<int>(local().size()); at_lvl++) {
+        const int ia1    = atomic_orbitals(at_lvl).first;
+        const auto& atom = ctx_.unit_cell().atom(ia1);
+        const int lo     = atomic_orbitals(at_lvl).second;
+        if (atom.type().lo_descriptor_hub(lo).use_for_calculation()) {
+            const int l      = atom.type().lo_descriptor_hub(lo).l();
+            const int offset = offset_[at_lvl];
             for (int m1 = -l; m1 <= l; m1++) {
                 for (int m2 = -l; m2 <= l; m2++) {
                     if (what__ == "get") {
                         for (int j = 0; j < ((ctx_.num_mag_dims() == 3) ? 4 : ctx_.num_spins()); j++) {
-                            occ_mtrx(l + m1, l + m2, j, ia) = this->local(ia)(l + m1, l + m2, j);
+                            occ_mtrx(offset + l + m1, offset + l + m2, j, ia1) = this->local(at_lvl)(l + m1, l + m2, j);
                         }
                     } else {
                         for (int j = 0; j < ((ctx_.num_mag_dims() == 3) ? 4 : ctx_.num_spins()); j++) {
-                            this->local(ia)(l + m1, l + m2, j) = occ_mtrx(l + m1, l + m2, j, ia);
+                            this->local(at_lvl)(l + m1, l + m2, j) = occ_mtrx(offset + l + m1, offset + l + m2, j, ia1);
                         }
                     }
                 }
@@ -95,36 +128,30 @@ void Hubbard_matrix::access(std::string const& what__, double_complex* occ__, in
     }
 }
 
-void Hubbard_matrix::print_local(int ia__, std::ostream& out__) const
+void
+Hubbard_matrix::print_local(int at_lvl__, std::ostream& out__) const
 {
-    auto const& atom = ctx_.unit_cell().atom(ia__);
-    if (!atom.type().hubbard_correction()) {
-        return;
-    }
     int const prec{5};
     int const width{10};
-    auto draw_bar = [&](int w)
-    {
-        out__ << std::setfill('-') << std::setw(w) << '-' << std::setfill(' ') << std::endl;
-    };
-    auto print_number = [&](double x)
-    {
-        out__ << std::setw(width) << std::setprecision(prec) << std::fixed << x;
-    };
 
-    out__ << "atom : " << ia__ << std::endl;
+    auto draw_bar = [&](int w) { out__ << std::setfill('-') << std::setw(w) << '-' << std::setfill(' ') << std::endl; };
+    auto print_number = [&](double x) { out__ << std::setw(width) << std::setprecision(prec) << std::fixed << x; };
+    auto const& atom  = ctx_.unit_cell().atom(atomic_orbitals_[at_lvl__].first);
+
+    out__ << "level : " << atom.type().lo_descriptor_hub(atomic_orbitals_[at_lvl__].second).n();
+    out__ << " l: " << atom.type().lo_descriptor_hub(atomic_orbitals_[at_lvl__].second).l() << std::endl;
+    const int l = atom.type().lo_descriptor_hub(atomic_orbitals_[at_lvl__].second).l();
     if (ctx_.num_mag_dims() != 3) {
-        int l = atom.type().indexr_hub().am(0).l();
-        int mmax = 2 *l + 1;
+        int mmax = 2 * l + 1;
         for (int is = 0; is < ctx_.num_spins(); is++) {
             draw_bar(width * mmax);
             bool has_imag{false};
             for (int m = 0; m < mmax; m++) {
                 for (int mp = 0; mp < mmax; mp++) {
-                    if (std::abs(std::imag(this->local(ia__)(m, mp, is))) > 1e-12) {
+                    if (std::abs(std::imag(this->local(at_lvl__)(m, mp, is))) > 1e-12) {
                         has_imag = true;
                     }
-                    print_number(std::real(this->local(ia__)(m, mp, is)));
+                    print_number(std::real(this->local(at_lvl__)(m, mp, is)));
                 }
                 out__ << std::endl;
             }
@@ -132,7 +159,7 @@ void Hubbard_matrix::print_local(int ia__, std::ostream& out__) const
                 out__ << "imaginary part:" << std::endl;
                 for (int m = 0; m < mmax; m++) {
                     for (int mp = 0; mp < mmax; mp++) {
-                        print_number(std::imag(this->local(ia__)(m, mp, is)));
+                        print_number(std::imag(this->local(at_lvl__)(m, mp, is)));
                     }
                     out__ << std::endl;
                 }
@@ -140,27 +167,26 @@ void Hubbard_matrix::print_local(int ia__, std::ostream& out__) const
         }
         draw_bar(width * mmax);
     } else {
-        int l = atom.type().indexr_hub().am(0).l();
-        int mmax = 2 *l + 1;
+        int mmax = 2 * l + 1;
         draw_bar(2 * width * mmax + 3);
         for (int m = 0; m < mmax; m++) {
             for (int mp = 0; mp < mmax; mp++) {
-                print_number(std::real(this->local(ia__)(m, mp, 0)));
+                print_number(std::real(this->local(at_lvl__)(m, mp, 0)));
             }
             out__ << " | ";
             for (int mp = 0; mp < mmax; mp++) {
-                print_number(std::real(this->local(ia__)(m, mp, 2)));
+                print_number(std::real(this->local(at_lvl__)(m, mp, 2)));
             }
             out__ << std::endl;
         }
         draw_bar(2 * width * mmax + 3);
         for (int m = 0; m < mmax; m++) {
             for (int mp = 0; mp < mmax; mp++) {
-                print_number(std::real(this->local(ia__)(m, mp, 3)));
+                print_number(std::real(this->local(at_lvl__)(m, mp, 3)));
             }
             out__ << " | ";
             for (int mp = 0; mp < mmax; mp++) {
-                print_number(std::real(this->local(ia__)(m, mp, 1)));
+                print_number(std::real(this->local(at_lvl__)(m, mp, 1)));
             }
             out__ << std::endl;
         }
@@ -168,35 +194,30 @@ void Hubbard_matrix::print_local(int ia__, std::ostream& out__) const
     }
 }
 
-void Hubbard_matrix::print_nonlocal(int idx__, std::ostream& out__) const
+void
+Hubbard_matrix::print_nonlocal(int idx__, std::ostream& out__) const
 {
 
-    auto nl = ctx_.cfg().hubbard().nonlocal(idx__);
-    int ia = nl.atom_pair()[0];
-    int ja = nl.atom_pair()[1];
-    int il = nl.l()[0];
-    int jl = nl.l()[1];
-    int ib = ctx_.unit_cell().atom(ia).type().indexr_hub().subshell_size(il, 0);
-    int jb = ctx_.unit_cell().atom(ja).type().indexr_hub().subshell_size(jl, 0);
-
+    auto nl      = ctx_.cfg().hubbard().nonlocal(idx__);
+    int ia       = nl.atom_pair()[0];
+    int ja       = nl.atom_pair()[1];
+    int il       = nl.l()[0];
+    int jl       = nl.l()[1];
+    const int jb = 2 * jl + 1;
+    const int ib = 2 * il + 1;
     vector3d<int> T(nl.T());
 
-    out__ << "atom: " << ia << ", l: " << il << " -> atom: " << ja << ", l: " << jl << ", T: " << T << std::endl;
+    vector3d<double> r = ctx_.unit_cell().atom(ja).position() + T - ctx_.unit_cell().atom(ia).position();
+    /* convert to Cartesian coordinates */
+    auto rc = dot(ctx_.unit_cell().lattice_vectors(), r);
+
+    out__ << "atom: " << ia << ", l: " << il << " -> atom: " << ja << ", l: " << jl << ", T: " << T
+          << ", r: " << rc << std::endl;
 
-    //auto const& atom = ctx_.unit_cell().atom(ia__);
-    //if (!atom.type().hubbard_correction()) {
-    //    return;
-    //}
     int const prec{5};
     int const width{10};
-    auto draw_bar = [&](int w)
-    {
-        out__ << std::setfill('-') << std::setw(w) << '-' << std::setfill(' ') << std::endl;
-    };
-    auto print_number = [&](double x)
-    {
-        out__ << std::setw(width) << std::setprecision(prec) << std::fixed << x;
-    };
+    auto draw_bar = [&](int w) { out__ << std::setfill('-') << std::setw(w) << '-' << std::setfill(' ') << std::endl; };
+    auto print_number = [&](double x) { out__ << std::setw(width) << std::setprecision(prec) << std::fixed << x; };
 
     if (ctx_.num_mag_dims() != 3) {
         for (int is = 0; is < ctx_.num_spins(); is++) {
@@ -252,17 +273,16 @@ void Hubbard_matrix::print_nonlocal(int idx__, std::ostream& out__) const
     //}
 }
 
-void Hubbard_matrix::zero()
+void
+Hubbard_matrix::zero()
 {
-    for (int ia = 0; ia < ctx_.unit_cell().num_atoms(); ia++) {
-        if (ctx_.unit_cell().atom(ia).type().hubbard_correction()) {
-            local_[ia].zero();
-        }
+    for (int ia = 0; ia < static_cast<int>(local_.size()); ia++) {
+        local_[ia].zero();
     }
-    for (int i = 0; i < ctx_.cfg().hubbard().nonlocal().size(); i++) {
+
+    for (int i = 0; i < static_cast<int>(ctx_.cfg().hubbard().nonlocal().size()); i++) {
         nonlocal_[i].zero();
     }
 }
 
-}
-
+} // namespace sirius
