@@ -32,70 +32,14 @@
 
 namespace sirius {
 
-class K_point_base // TODO: good name? maybe k_point?
-{
-  protected:
-    /// Fractional k-point coordinates.
-    std::array<double, 3> vk_;
-
-    /// List of G-vectors with |G+k| < cutoff.
-    std::unique_ptr<Gvec> gkvec_;
-
-    /// Communicator for parallelization inside k-point.
-    /** This communicator is used to split G+k vectors and wave-functions. */
-    sddk::Communicator const& comm_;
-
-    void init(double gk_cutoff__, matrix3d<double> reciprocal_lattice_vectors__, bool gamma__)
-    {
-        /* create G+k vectors; communicator of the coarse FFT grid is used because wave-functions will be transformed
-         * only on the coarse grid; G+k-vectors will be distributed between MPI ranks assigned to the k-point */
-        gkvec_ = std::unique_ptr<Gvec>(new Gvec(vk_, reciprocal_lattice_vectors__, gk_cutoff__, comm_, gamma__));
-    }
-
-  public:
-    K_point_base(std::array<double, 3> vk__, double gk_cutoff__, matrix3d<double> reciprocal_lattice_vectors__,
-                 bool gamma__)
-        : vk_(vk__)
-        , comm_(sddk::Communicator::self())
-    {
-        init(gk_cutoff__, reciprocal_lattice_vectors__, gamma__);
-    }
-
-    K_point_base(std::array<double, 3> vk__, double gk_cutoff__, matrix3d<double> reciprocal_lattice_vectors__,
-                 bool gamma__, sddk::Communicator const& comm__)
-        : vk_(vk__)
-        , comm_(comm__)
-    {
-        init(gk_cutoff__, reciprocal_lattice_vectors__, gamma__);
-    }
-
-    inline Gvec const& gkvec() const
-    {
-        return *gkvec_;
-    }
-
-    /// Total number of G+k vectors within the cutoff distance
-    inline int num_gkvec() const
-    {
-        return gkvec_->num_gvec();
-    }
-
-    /// Local number of G+k vectors in case of flat distribution.
-    inline int num_gkvec_loc() const
-    {
-        return gkvec().count();
-    }
-
-};
-
 /// K-point related variables and methods.
 /** \image html wf_storage.png "Wave-function storage"
  *  \image html fv_eigen_vectors.png "First-variational eigen vectors"
  *
- *   \tparam T  Precision of the wave-functions (float or double).
+ *  \tparam T  Precision of the wave-functions (float or double).
  */
 template <typename T>
-class K_point : public K_point_base
+class K_point
 {
   private:
     /// Simulation context.
@@ -104,12 +48,23 @@ class K_point : public K_point_base
     /// Unit cell object.
     Unit_cell const& unit_cell_;
 
-    /// K-point id.
-    int id_{-1};
+    /// Fractional k-point coordinates.
+    vector3d<double> vk_;
 
     /// Weight of k-point.
     double weight_{1.0};
 
+    /// Communicator for parallelization inside k-point.
+    /** This communicator is used to split G+k vectors and wave-functions. */
+    sddk::Communicator const& comm_;
+
+    /// List of G-vectors with |G+k| < cutoff.
+    std::shared_ptr<Gvec> gkvec_;
+
+    std::shared_ptr<Gvec> gkvec_row_;
+
+    std::shared_ptr<Gvec> gkvec_col_;
+
     /// G-vector distribution for the FFT transformation.
     std::unique_ptr<Gvec_partition> gkvec_partition_;
 
@@ -142,19 +97,17 @@ class K_point : public K_point_base
     /// Two-component (spinor) wave functions describing the bands.
     std::shared_ptr<Wave_functions<T>> spinor_wave_functions_{nullptr};
 
-    /// Two-component (spinor) wave functions used to compute the Hubbard corrections.
-    /** This wave-functions are not necessarily equal to atomic wave-functions. */
-    std::unique_ptr<Wave_functions<T>> wave_functions_hub_{nullptr};
+    /// Pseudopotential atmoic wave-functions (not orthogonalized).
+    std::unique_ptr<Wave_functions<T>> atomic_wave_functions_{nullptr};
 
-    /// Two-component (spinor) wave functions used to compute the Hubbard corrections.
-    /** This wave-functions are not necessarily equal to atomic wave-functions. S-operator is applied to this WFs. */
-    std::unique_ptr<Wave_functions<T>> wave_functions_S_hub_{nullptr};
+    /// Pseudopotential atmoic wave-functions (not orthogonalized) with S-operator applied.
+    std::unique_ptr<Wave_functions<T>> atomic_wave_functions_S_{nullptr};
 
-    /// Two-component (spinor) atomic orbitals used to compute the Hubbard wave functions
-    std::unique_ptr<Wave_functions<T>> atomic_wave_functions_hub_{nullptr};
+    /// Hubbard wave functions.
+    std::unique_ptr<Wave_functions<T>> hubbard_wave_functions_{nullptr};
 
-    /// Two-component (spinor) atomic orbitals (with the S operator applied for uspp) used to compute the Hubbard wave functions
-    std::unique_ptr<Wave_functions<T>> atomic_wave_functions_S_hub_{nullptr};
+    /// Hubbard wave functions with S-operator applied.
+    std::unique_ptr<Wave_functions<T>> hubbard_wave_functions_S_{nullptr};
 
     /// Band occupation numbers.
     sddk::mdarray<double, 2> band_occupancies_;
@@ -191,9 +144,6 @@ class K_point : public K_point_base
     /// Number of G+k vectors distributed along columns of MPI grid
     int num_gkvec_col_{0};
 
-    /// Offset of the local fraction of G+k vectors in the global index.
-    int gkvec_offset_{0};
-
     /// Basis descriptors distributed between rows of the 2D MPI grid.
     /** This is a local array. Only MPI ranks belonging to the same column have identical copies of this array. */
     std::vector<lo_basis_descriptor> lo_basis_descriptors_row_;
@@ -259,30 +209,14 @@ class K_point : public K_point_base
 
     inline int get_ispn(int ispn__) const
     {
-        assert(ispn__ == 0 || ispn__ == 1);
+        RTE_ASSERT(ispn__ == 0 || ispn__ == 1);
         return ispn_map_[ispn__];
     }
 
     friend class K_point_set;
 
-  public:
-    /// Constructor
-    K_point(Simulation_context& ctx__, double const* vk__, double weight__, int id__)
-        : K_point_base(std::array<double, 3>({vk__[0], vk__[1], vk__[2]}), ctx__.gk_cutoff(),
-                       ctx__.unit_cell().reciprocal_lattice_vectors(), ctx__.gamma_point(), ctx__.comm_band())
-        , ctx_(ctx__)
-        , unit_cell_(ctx_.unit_cell())
-        , id_(id__)
-        , weight_(weight__)
-        , comm_row_(ctx_.blacs_grid().comm_row())
-        , comm_col_(ctx_.blacs_grid().comm_col())
+    void init0()
     {
-        PROFILE("sirius::K_point::K_point");
-
-        for (int x = 0; x < 3; x++) {
-            vk_[x] = vk__[x];
-        }
-
         band_occupancies_ = sddk::mdarray<double, 2>(ctx_.num_bands(), ctx_.num_spinors(),
                                                      memory_t::host, "band_occupancies");
         band_occupancies_.zero();
@@ -290,12 +224,6 @@ class K_point : public K_point_base
                                                   memory_t::host, "band_energies");
         band_energies_.zero();
 
-        num_ranks_row_ = comm_row_.size();
-        num_ranks_col_ = comm_col_.size();
-
-        rank_row_ = comm_row_.rank();
-        rank_col_ = comm_col_.rank();
-
         ispn_map_[0] = 0;
         ispn_map_[1] = -1;
         if (ctx_.num_mag_dims() == 1) {
@@ -305,6 +233,44 @@ class K_point : public K_point_base
         }
     }
 
+  public:
+    /// Constructor
+    K_point(Simulation_context& ctx__, vector3d<double> vk__, double weight__)
+        : ctx_(ctx__)
+        , unit_cell_(ctx_.unit_cell())
+        , vk_(vk__)
+        , weight_(weight__)
+        , comm_(ctx_.comm_band())
+        , rank_col_(ctx_.blacs_grid().comm_col().rank())
+        , num_ranks_col_(ctx_.blacs_grid().comm_col().size())
+        , rank_row_(ctx_.blacs_grid().comm_row().rank())
+        , num_ranks_row_(ctx_.blacs_grid().comm_row().size())
+        , comm_row_(ctx_.blacs_grid().comm_row())
+        , comm_col_(ctx_.blacs_grid().comm_col())
+    {
+        this->init0();
+        gkvec_ = std::make_shared<Gvec>(vk_, unit_cell_.reciprocal_lattice_vectors(), ctx_.gk_cutoff(), comm_,
+                                        ctx_.gamma_point());
+    }
+
+    /// Constructor
+    K_point(Simulation_context& ctx__, std::shared_ptr<Gvec> gkvec__, double weight__)
+        : ctx_(ctx__)
+        , unit_cell_(ctx_.unit_cell())
+        , vk_(gkvec__->vk())
+        , weight_(weight__)
+        , comm_(ctx_.comm_band())
+        , gkvec_(gkvec__)
+        , rank_col_(ctx_.blacs_grid().comm_col().rank())
+        , num_ranks_col_(ctx_.blacs_grid().comm_col().size())
+        , rank_row_(ctx_.blacs_grid().comm_row().rank())
+        , num_ranks_row_(ctx_.blacs_grid().comm_row().size())
+        , comm_row_(ctx_.blacs_grid().comm_row())
+        , comm_col_(ctx_.blacs_grid().comm_col())
+    {
+        this->init0();
+    }
+
     /// Initialize the k-point related arrays and data.
     void initialize(); // TODO: initialize from HDF5
 
@@ -379,37 +345,11 @@ class K_point : public K_point_base
      */
     void generate_atomic_wave_functions(std::vector<int> atoms__,
                                         std::function<sirius::experimental::basis_functions_index const*(int)> indexb__,
-                                        Radial_integrals_atomic_wf<false> const& ri__, sddk::Wave_functions<T>& wf__);
-
-    void compute_gradient_wave_functions(Wave_functions<T>& phi, const int starting_position_i, const int num_wf,
-                                         Wave_functions<T>& dphi, const int starting_position_j, const int direction);
+                                        Radial_integrals_atomic_wf<false> const& ri__,
+                                        sddk::Wave_functions<T>& wf__);
 
     void generate_hubbard_orbitals();
 
-    void copy_hubbard_orbitals_on_device()
-    {
-        if (ctx_.hubbard_correction() && is_device_memory(ctx_.preferred_memory_t())) {
-            auto& mpd = ctx_.mem_pool(memory_t::device);
-            for (int ispn = 0; ispn < this->wave_functions_S_hub().num_sc(); ispn++) {
-                this->wave_functions_S_hub().pw_coeffs(ispn).allocate(mpd);
-                this->wave_functions_S_hub().pw_coeffs(ispn).copy_to(memory_t::device, 0,
-                        this->wave_functions_S_hub().num_wf());
-            }
-        }
-    }
-
-    void release_hubbard_orbitals_on_device()
-    {
-        if (ctx_.hubbard_correction() && is_device_memory(ctx_.preferred_memory_t())) {
-            for (int ispn = 0; ispn < this->wave_functions_S_hub().num_sc(); ispn++) {
-                this->wave_functions_S_hub().pw_coeffs(ispn).deallocate(memory_t::device);
-            }
-        }
-    }
-
-    void orthogonalize_hubbard_orbitals(Wave_functions<T>& phi__, Wave_functions<T>& sphi__,
-                                        Wave_functions<T>& phi_hub__, Wave_functions<T>& sphi_hub__);
-
     /// Save data to HDF5 file.
     void save(std::string const& name__, int id__) const;
 
@@ -422,11 +362,14 @@ class K_point : public K_point_base
     /// Collect distributed first-variational vectors into a global array.
     void get_fv_eigen_vectors(mdarray<std::complex<T>, 2>& fv_evec__) const;
 
+    /// Test orthonormalization of spinor wave-functions
+    void test_spinor_wave_functions(int use_fft);
+
     /// Collect distributed second-variational vectors into a global array.
     void get_sv_eigen_vectors(mdarray<std::complex<T>, 2>& sv_evec__) const
     {
-        assert((int)sv_evec__.size(0) == ctx_.num_spins() * ctx_.num_fv_states());
-        assert((int)sv_evec__.size(1) == ctx_.num_spins() * ctx_.num_fv_states());
+        RTE_ASSERT((int)sv_evec__.size(0) == ctx_.num_spins() * ctx_.num_fv_states());
+        RTE_ASSERT((int)sv_evec__.size(1) == ctx_.num_spins() * ctx_.num_fv_states());
 
         sv_evec__.zero();
 
@@ -444,7 +387,7 @@ class K_point : public K_point_base
             for (int jloc = 0; jloc < sv_eigen_vectors_[ispn].num_cols_local(); jloc++) {
                 int j = sv_eigen_vectors_[ispn].icol(jloc);
                 for (int iloc = 0; iloc < sv_eigen_vectors_[ispn].num_rows_local(); iloc++) {
-                    int i = sv_eigen_vectors_[ispn].irow(iloc);
+                    int i                         = sv_eigen_vectors_[ispn].irow(iloc);
                     sv_evec__(i + offs, j + offs) = sv_eigen_vectors_[ispn](iloc, jloc);
                 }
             }
@@ -453,8 +396,22 @@ class K_point : public K_point_base
         comm().allreduce(sv_evec__.at(memory_t::host), (int)sv_evec__.size());
     }
 
-    /// Test orthonormalization of spinor wave-functions
-    void test_spinor_wave_functions(int use_fft);
+    inline Gvec const& gkvec() const
+    {
+        return *gkvec_;
+    }
+
+    /// Total number of G+k vectors within the cutoff distance
+    inline int num_gkvec() const
+    {
+        return gkvec_->num_gvec();
+    }
+
+    /// Local number of G+k vectors in case of flat distribution.
+    inline int num_gkvec_loc() const
+    {
+        return gkvec().count();
+    }
 
     /// Get the number of occupied bands for each spin channel.
     inline int num_occupied_bands(int ispn__ = -1) const
@@ -527,54 +484,56 @@ class K_point : public K_point_base
         return spinor_wave_functions_;
     }
 
-   // the S operator is applied on these functions
-    inline Wave_functions<T> const& wave_functions_S_hub() const
+    /// return the initial atomic orbitals used to compute the hubbard wave functions. the S operator is applied on
+    /// these functions
+    inline Wave_functions<T> const& atomic_wave_functions_S() const
     {
         /* the S operator is applied on these functions */
-        assert(wave_functions_S_hub_ != nullptr);
-        return *wave_functions_S_hub_;
+        RTE_ASSERT(atomic_wave_functions_S_ != nullptr);
+        return *atomic_wave_functions_S_;
     }
 
-    inline Wave_functions<T>& wave_functions_S_hub()
+    inline Wave_functions<T>& atomic_wave_functions_S()
     {
-        return const_cast<Wave_functions<T>&>(static_cast<K_point const&>(*this).wave_functions_S_hub());
+        return const_cast<Wave_functions<T>&>(static_cast<K_point const&>(*this).atomic_wave_functions_S());
     }
 
-    inline Wave_functions<T> const& wave_functions_hub() const
+    inline Wave_functions<T> const& atomic_wave_functions() const
     {
-        assert(wave_functions_hub_!= nullptr);
-        return *wave_functions_hub_;
+        RTE_ASSERT(atomic_wave_functions_ != nullptr);
+        return *atomic_wave_functions_;
     }
 
-    inline Wave_functions<T>& wave_functions_hub()
+    /// return the initial atomic orbitals used to compute the hubbard wave functions.
+    inline Wave_functions<T>& atomic_wave_functions()
     {
-        return const_cast<Wave_functions<T>&>(static_cast<K_point const&>(*this).wave_functions_hub());
+        return const_cast<Wave_functions<T>&>(static_cast<K_point const&>(*this).atomic_wave_functions());
     }
 
-    /// Return the atomic wave functions used to compute the hubbard wave functions. The S operator is applied when uspp are used.
-    inline Wave_functions<T> const& atomic_wave_functions_S_hub() const
+    /// return the actual hubbard wave functions used in the calculations. The S operator is applied when uspp are used.
+    inline Wave_functions<T> const& hubbard_wave_functions_S() const
     {
-        assert(atomic_wave_functions_S_hub_ != nullptr);
-        return *atomic_wave_functions_S_hub_;
+        assert(hubbard_wave_functions_S_ != nullptr);
+        return *hubbard_wave_functions_S_;
     }
 
-    /// return the atomic wave functions used to compute the hubbard wave functions. The S operator is applied when uspp are used.
-    inline Wave_functions<T>& atomic_wave_functions_S_hub()
+    /// return the actual hubbard wave functions used in the calculations. The S operator is applied when uspp are used.
+    inline Wave_functions<T>& hubbard_wave_functions_S()
     {
-        return const_cast<Wave_functions<T>&>(static_cast<K_point const&>(*this).atomic_wave_functions_S_hub());
+        return const_cast<Wave_functions<T>&>(static_cast<K_point const&>(*this).hubbard_wave_functions_S());
     }
 
-    /// return the atomic wave functions used to compute the hubbard wave functions.
-    inline Wave_functions<T> const& atomic_wave_functions_hub() const
+    /// return the actual hubbard wave functions used in the calculations.
+    inline Wave_functions<T> const& hubbard_wave_functions() const
     {
-        assert(atomic_wave_functions_hub_ != nullptr);
-        return *atomic_wave_functions_hub_;
+        assert(hubbard_wave_functions_ != nullptr);
+        return *hubbard_wave_functions_;
     }
 
-    /// return the atomic wave functions used to compute the hubbard wave functions.
-    inline Wave_functions<T>& atomic_wave_functions_hub()
+    /// return the actual hubbard wave functions used in the calculations.
+    inline Wave_functions<T>& hubbard_wave_functions()
     {
-        return const_cast<Wave_functions<T>&>(static_cast<K_point const&>(*this).atomic_wave_functions_hub());
+        return const_cast<Wave_functions<T>&>(static_cast<K_point const&>(*this).hubbard_wave_functions());
     }
 
     inline Wave_functions<T>& singular_components()
@@ -595,10 +554,10 @@ class K_point : public K_point_base
         return num_gkvec() + unit_cell_.mt_lo_basis_size();
     }
 
-    /// Return global index of G+k vector.
+    /// Return global index of G+k vector. // TODO: consider removal
     inline int idxgk(int igkloc__) const
     {
-        return gkvec_offset_ + igkloc__;
+        return gkvec_->offset() + igkloc__;
     }
 
     /// Local number of G+k vectors for each MPI rank in the row of the 2D MPI grid.
