@@ -49,11 +49,11 @@ namespace sirius {
     For the function expanded in plane-waves we have:
     \f[
       f_{\mathrm{sym}}({\bf r}) = \frac{1}{N_{\mathrm{sym}}}
-        \sum_{\hat{\bf S}\hat{\bf P}} \hat {\bf S} \hat {\bf P}f({\bf r}) = 
+        \sum_{\hat{\bf S}\hat{\bf P}} \hat {\bf S} \hat {\bf P}f({\bf r}) =
         \frac{1}{N_{\mathrm{sym}}} \sum_{\hat{\bf S}\hat{\bf P}} \hat {\bf S} \sum_{\bf G}
         f({\bf G}) e^{i{\bf G}\hat{\bf P}^{-1}{\bf r}} = \\
         \frac{1}{N_{\mathrm{sym}}} \sum_{\hat{\bf S}\hat{\bf P}} \sum_{\bf G} \hat {\bf S} f({\bf G})
-        e^{i{\bf G}({\bf R}^{-1}{\bf r} - {\bf R}^{-1}{\bf t})} = 
+        e^{i{\bf G}({\bf R}^{-1}{\bf r} - {\bf R}^{-1}{\bf t})} =
         \frac{1}{N_{\mathrm{sym}}} \sum_{\hat{\bf S}\hat{\bf P}} \sum_{\bf G} \hat {\bf S} f({\bf G})
         e^{i{\bf G}'{\bf r}} e^{-i{\bf G}'{\bf t}}
     \f]
@@ -62,7 +62,7 @@ namespace sirius {
     rewrite the expression using inverse relation \f$ {\bf G} = {\bf R}^{T}{\bf G'} \f$ and summing over <b>G</b>'
     (which is just a permutaion of <b>G</b>):
     \f[
-       f_{\mathrm{sym}}({\bf r}) = 
+       f_{\mathrm{sym}}({\bf r}) =
         \sum_{\bf G'} e^{i{\bf G}'{\bf r}} \frac{1}{N_{\mathrm{sym}}} \sum_{\hat{\bf S}\hat{\bf P}}
         \hat {\bf S} f({\bf R}^{T}{\bf G'}) e^{-i{\bf G}'{\bf t}}
     \f]
@@ -81,11 +81,11 @@ namespace sirius {
     which leads to the following relation for the plane-wave coefficient:
     \f[
       \sum_{\bf G} f_{\mathrm{sym}}({\bf G})e^{i{\bf G}{\bf r}} =
-        \sum_{\bf G} \hat{\bf S}f_{\mathrm{sym}}({\bf G})e^{i{\bf G}\hat{\bf P}^{-1}{\bf r}} = 
+        \sum_{\bf G} \hat{\bf S}f_{\mathrm{sym}}({\bf G})e^{i{\bf G}\hat{\bf P}^{-1}{\bf r}} =
         \sum_{\bf G} \hat{\bf S}f_{\mathrm{sym}}({\bf G})e^{i{\bf G}{\bf R}^{-1}{\bf r}}
-            e^{-i{\bf G}{\bf R}^{-1}{\bf t}} = 
+            e^{-i{\bf G}{\bf R}^{-1}{\bf t}} =
         \sum_{\bf G'} \hat{\bf S}f_{\mathrm{sym}}({\bf G})e^{i{\bf G}'{\bf r}}
-             e^{-i{\bf G}'{\bf t}} = 
+             e^{-i{\bf G}'{\bf t}} =
         \sum_{\bf G'} \hat{\bf S}f_{\mathrm{sym}}({\bf G'})e^{i{\bf G}'{\bf r}}
     \f]
     and so
@@ -93,9 +93,10 @@ namespace sirius {
        f_{\mathrm{sym}}({\bf G}') = \hat{\bf S}f_{\mathrm{sym}}({\bf G})e^{-i{\bf G'}{\bf t}}
     \f]
  */
-inline void symmetrize(Crystal_symmetry const& sym__, Gvec_shells const& gvec_shells__,
-                       sddk::mdarray<double_complex, 3> const& sym_phase_factors__, double_complex* f_pw__,
-                       double_complex* x_pw__, double_complex* y_pw__, double_complex* z_pw__)
+inline void
+symmetrize(Crystal_symmetry const& sym__, Gvec_shells const& gvec_shells__,
+           sddk::mdarray<double_complex, 3> const& sym_phase_factors__, double_complex* f_pw__, double_complex* x_pw__,
+           double_complex* y_pw__, double_complex* z_pw__)
 {
     PROFILE("sirius::symmetrize|fpw");
 
@@ -118,10 +119,8 @@ inline void symmetrize(Crystal_symmetry const& sym__, Gvec_shells const& gvec_sh
 
     double norm = 1 / double(sym__.size());
 
-    auto phase_factor = [&](int isym, vector3d<int> G)
-    {
-        return sym_phase_factors__(0, G[0], isym) *
-               sym_phase_factors__(1, G[1], isym) *
+    auto phase_factor = [&](int isym, vector3d<int> G) {
+        return sym_phase_factors__(0, G[0], isym) * sym_phase_factors__(1, G[1], isym) *
                sym_phase_factors__(2, G[2], isym);
     };
 
@@ -131,7 +130,7 @@ inline void symmetrize(Crystal_symmetry const& sym__, Gvec_shells const& gvec_sh
 
     #pragma omp parallel
     {
-        int nt = omp_get_max_threads();
+        int nt  = omp_get_max_threads();
         int tid = omp_get_thread_num();
 
         for (int igloc = 0; igloc < gvec_shells__.gvec_count_remapped(); igloc++) {
@@ -233,9 +232,9 @@ inline void symmetrize(Crystal_symmetry const& sym__, Gvec_shells const& gvec_sh
                         }
                         if (is_non_collin) {
                             auto v = dot(S, vector3d<double_complex>({symx, symy, symz}));
-                            symx1 = v[0] * phase;
-                            symy1 = v[1] * phase;
-                            symz1 = v[2] * phase;
+                            symx1  = v[0] * phase;
+                            symy1  = v[1] * phase;
+                            symz1  = v[2] * phase;
                         }
 
                         if (is_done[ig1]) {
@@ -279,7 +278,6 @@ inline void symmetrize(Crystal_symmetry const& sym__, Gvec_shells const& gvec_sh
                                 sym_x_pw[ig1] = symx1;
                                 sym_y_pw[ig1] = symy1;
                                 sym_z_pw[ig1] = symz1;
-
                             }
                             is_done[ig1] = true;
                         }
@@ -294,10 +292,10 @@ inline void symmetrize(Crystal_symmetry const& sym__, Gvec_shells const& gvec_sh
     for (int igloc = 0; igloc < gvec_shells__.gvec_count_remapped(); igloc++) {
         auto G = gvec_shells__.gvec_remapped(igloc);
         for (int isym = 0; isym < sym__.size(); isym++) {
-            auto S = sym__[isym].spin_rotation;
+            auto S      = sym__[isym].spin_rotation;
             auto gv_rot = dot(sym__[isym].spg_op.invRT, G);
             /* index of a rotated G-vector */
-            int ig_rot = gvec_shells__.index_by_gvec(gv_rot);
+            int ig_rot           = gvec_shells__.index_by_gvec(gv_rot);
             double_complex phase = std::conj(phase_factor(isym, gv_rot));
 
             if (f_pw__ && ig_rot != -1) {
@@ -334,7 +332,8 @@ inline void symmetrize(Crystal_symmetry const& sym__, Gvec_shells const& gvec_sh
     }
 }
 
-inline void symmetrize_function(Crystal_symmetry const& sym__, Communicator const& comm__, mdarray<double, 3>& frlm__)
+inline void
+symmetrize_function(Crystal_symmetry const& sym__, Communicator const& comm__, mdarray<double, 3>& frlm__)
 {
     PROFILE("sirius::symmetrize_function|flm");
 
@@ -357,25 +356,26 @@ inline void symmetrize_function(Crystal_symmetry const& sym__, Communicator cons
 
     for (int i = 0; i < sym__.size(); i++) {
         /* full space-group symmetry operation is {R|t} */
-        int pr = sym__[i].spg_op.proper;
+        int pr    = sym__[i].spg_op.proper;
         auto eang = sym__[i].spg_op.euler_angles;
         sht::rotation_matrix(lmax, eang, pr, rotm);
 
         for (int ialoc = 0; ialoc < spl_atoms.local_size(); ialoc++) {
             int ia = spl_atoms[ialoc];
             int ja = sym__[i].spg_op.inv_sym_atom[ia];
-            linalg(linalg_t::blas).gemm('N', 'N', lmmax, nrmax, lmmax, &alpha, rotm.at(memory_t::host), rotm.ld(),
-                                        frlm__.at(memory_t::host, 0, 0, ja), frlm__.ld(), &linalg_const<double>::one(),
-                                        fsym.at(memory_t::host, 0, 0, ialoc), fsym.ld());
+            linalg(linalg_t::blas)
+                .gemm('N', 'N', lmmax, nrmax, lmmax, &alpha, rotm.at(memory_t::host), rotm.ld(),
+                      frlm__.at(memory_t::host, 0, 0, ja), frlm__.ld(), &linalg_const<double>::one(),
+                      fsym.at(memory_t::host, 0, 0, ialoc), fsym.ld());
         }
     }
     double* sbuf = spl_atoms.local_size() ? fsym.at(memory_t::host) : nullptr;
     comm__.allgather(sbuf, frlm__.at(memory_t::host), lmmax * nrmax * spl_atoms.local_size(),
-            lmmax * nrmax * spl_atoms.global_offset());
+                     lmmax * nrmax * spl_atoms.global_offset());
 }
 
-inline void symmetrize_vector_function(Crystal_symmetry const& sym__, Communicator const& comm__,
-                                       mdarray<double, 3>& vz_rlm__)
+inline void
+symmetrize_vector_function(Crystal_symmetry const& sym__, Communicator const& comm__, mdarray<double, 3>& vz_rlm__)
 {
     PROFILE("sirius::symmetrize_function|vzlm");
 
@@ -399,29 +399,30 @@ inline void symmetrize_vector_function(Crystal_symmetry const& sym__, Communicat
 
     for (int i = 0; i < sym__.size(); i++) {
         /* full space-group symmetry operation is {R|t} */
-        int pr = sym__[i].spg_op.proper;
+        int pr    = sym__[i].spg_op.proper;
         auto eang = sym__[i].spg_op.euler_angles;
-        auto S = sym__[i].spin_rotation;
+        auto S    = sym__[i].spin_rotation;
         sht::rotation_matrix(lmax, eang, pr, rotm);
 
         for (int ialoc = 0; ialoc < spl_atoms.local_size(); ialoc++) {
-            int ia = spl_atoms[ialoc];
-            int ja = sym__[i].spg_op.inv_sym_atom[ia];
+            int ia   = spl_atoms[ialoc];
+            int ja   = sym__[i].spg_op.inv_sym_atom[ia];
             double a = alpha * S(2, 2);
-            linalg(linalg_t::blas).gemm('N', 'N', lmmax, nrmax, lmmax, &a, rotm.at(memory_t::host),
-            rotm.ld(), vz_rlm__.at(memory_t::host, 0, 0, ja), vz_rlm__.ld(), &linalg_const<double>::one(),
-            fsym.at(memory_t::host, 0, 0, ialoc), fsym.ld());
+            linalg(linalg_t::blas)
+                .gemm('N', 'N', lmmax, nrmax, lmmax, &a, rotm.at(memory_t::host), rotm.ld(),
+                      vz_rlm__.at(memory_t::host, 0, 0, ja), vz_rlm__.ld(), &linalg_const<double>::one(),
+                      fsym.at(memory_t::host, 0, 0, ialoc), fsym.ld());
         }
     }
 
     double* sbuf = spl_atoms.local_size() ? fsym.at(memory_t::host) : nullptr;
     comm__.allgather(sbuf, vz_rlm__.at(memory_t::host), lmmax * nrmax * spl_atoms.local_size(),
-            lmmax * nrmax * spl_atoms.global_offset());
+                     lmmax * nrmax * spl_atoms.global_offset());
 }
 
-inline void symmetrize_vector_function(Crystal_symmetry const& sym__, Communicator const& comm__,
-                                       mdarray<double, 3>& vx_rlm__, mdarray<double, 3>& vy_rlm__,
-                                       mdarray<double, 3>& vz_rlm__)
+inline void
+symmetrize_vector_function(Crystal_symmetry const& sym__, Communicator const& comm__, mdarray<double, 3>& vx_rlm__,
+                           mdarray<double, 3>& vy_rlm__, mdarray<double, 3>& vz_rlm__)
 {
     PROFILE("sirius::symmetrize_function|vlm");
 
@@ -445,22 +446,23 @@ inline void symmetrize_vector_function(Crystal_symmetry const& sym__, Communicat
 
     for (int i = 0; i < sym__.size(); i++) {
         /* full space-group symmetry operation is {R|t} */
-        int pr = sym__[i].spg_op.proper;
+        int pr    = sym__[i].spg_op.proper;
         auto eang = sym__[i].spg_op.euler_angles;
-        auto S = sym__[i].spin_rotation;
+        auto S    = sym__[i].spin_rotation;
         sht::rotation_matrix(lmax, eang, pr, rotm);
 
         for (int ialoc = 0; ialoc < spl_atoms.local_size(); ialoc++) {
             int ia = spl_atoms[ialoc];
             int ja = sym__[i].spg_op.inv_sym_atom[ia];
-            for (int k: {0, 1, 2}) {
-                linalg(linalg_t::blas).gemm('N', 'N', lmmax, nrmax, lmmax, &alpha, rotm.at(memory_t::host), rotm.ld(),
-                                            vrlm[k]->at(memory_t::host, 0, 0, ja), vrlm[k]->ld(),
-                                            &linalg_const<double>::zero(), vtmp.at(memory_t::host, 0, 0, k), vtmp.ld());
+            for (int k : {0, 1, 2}) {
+                linalg(linalg_t::blas)
+                    .gemm('N', 'N', lmmax, nrmax, lmmax, &alpha, rotm.at(memory_t::host), rotm.ld(),
+                          vrlm[k]->at(memory_t::host, 0, 0, ja), vrlm[k]->ld(), &linalg_const<double>::zero(),
+                          vtmp.at(memory_t::host, 0, 0, k), vtmp.ld());
             }
             #pragma omp parallel
-            for (int k: {0, 1, 2}) {
-                for (int j: {0, 1, 2}) {
+            for (int k : {0, 1, 2}) {
+                for (int j : {0, 1, 2}) {
                     #pragma omp for
                     for (int ir = 0; ir < nrmax; ir++) {
                         for (int lm = 0; lm < lmmax; lm++) {
@@ -472,10 +474,10 @@ inline void symmetrize_vector_function(Crystal_symmetry const& sym__, Communicat
         }
     }
 
-    for (int k: {0, 1, 2}) {
+    for (int k : {0, 1, 2}) {
         double* sbuf = spl_atoms.local_size() ? v_sym.at(memory_t::host, 0, 0, 0, k) : nullptr;
         comm__.allgather(sbuf, vrlm[k]->at(memory_t::host), lmmax * nrmax * spl_atoms.local_size(),
-                lmmax * nrmax * spl_atoms.global_offset());
+                         lmmax * nrmax * spl_atoms.global_offset());
     }
 }
 
@@ -484,7 +486,7 @@ inline void symmetrize_vector_function(Crystal_symmetry const& sym__, Communicat
 
 /* we compute \f[ Oc = (R_l\cross R_s) . O . (R_l\cross R_s)^\dagger \f] */
 
-//inline void apply_symmetry(const mdarray<double_complex, 3> &dm_,
+// inline void apply_symmetry(const mdarray<double_complex, 3> &dm_,
 //                           const mdarray<double, 2> &rot,
 //                           const mdarray<double_complex, 2> &spin_rot_su2,
 //                           const int num_mag_dims_,
@@ -502,7 +504,8 @@ inline void symmetrize_vector_function(Crystal_symmetry const& sym__, Communicat
 //                dm_rot_spatial[j] = 0.0;
 //                for (int lm3 = 0; lm3 <= 2 * l + 1; lm3++) {
 //                    for (int lm4 = 0; lm4 <= 2 * l + 1; lm4++) {
-//                        dm_rot_spatial[j] += dm_(lm3, lm4, j) * rot(l * l + lm1, l * l + lm3) * rot(l * l + lm2, l * l + lm4);
+//                        dm_rot_spatial[j] += dm_(lm3, lm4, j) * rot(l * l + lm1, l * l + lm3) * rot(l * l + lm2, l * l
+//                        + lm4);
 //                    }
 //                }
 //            }
@@ -524,7 +527,8 @@ inline void symmetrize_vector_function(Crystal_symmetry const& sym__, Communicat
 //                for (int k = 0; k < num_mag_dims_; k++) {
 //                    for (int is = 0; is < 2; is++) {
 //                        for (int js = 0; js < 2; js++) {
-//                            res_(lm1, lm2, k) += spin_rot_su2(k & 1, is) * spin_dm[is][js] * std::conj(spin_rot_su2(std::min(k, 1), js));
+//                            res_(lm1, lm2, k) += spin_rot_su2(k & 1, is) * spin_dm[is][js] *
+//                            std::conj(spin_rot_su2(std::min(k, 1), js));
 //                        }
 //                    }
 //                }
@@ -533,7 +537,6 @@ inline void symmetrize_vector_function(Crystal_symmetry const& sym__, Communicat
 //    }
 //}
 
-
 /// Symmetrize density or occupancy matrix according to a given list of basis functions.
 /** Density matrix arises in LAPW or PW methods. In PW it is computed in the basis of beta-projectors. Occupancy
  *  matrix is computed for the Hubbard-U correction. In both cases the matrix has the same structure and is
@@ -542,29 +545,24 @@ inline void symmetrize_vector_function(Crystal_symmetry const& sym__, Communicat
  *  that case, we must skip half of the indices because of the averaging of the
  *  radial integrals over the total angular momentum
  */
-inline void symmetrize(const mdarray<double_complex, 4> &ns_,
-                       const basis_functions_index &indexb,
-                       const int ia,
-                       const int ja,
-                       const int ndm,
-                       const mdarray<double, 2> &rotm,
-                       const mdarray<double_complex, 2> &spin_rot_su2,
-                       mdarray<double_complex, 4> &dm_,
-                       const bool hubbard_)
+inline void
+symmetrize(const mdarray<double_complex, 4>& ns_, const basis_functions_index& indexb, const int ia, const int ja,
+           const int ndm, const mdarray<double, 2>& rotm, const mdarray<double_complex, 2>& spin_rot_su2,
+           mdarray<double_complex, 4>& dm_, const bool hubbard_)
 {
     for (int xi1 = 0; xi1 < indexb.size(); xi1++) {
         int l1  = indexb[xi1].l;
         int lm1 = indexb[xi1].lm;
         int o1  = indexb[xi1].order;
 
-        if ((hubbard_)&&(xi1 >= (2 * l1 + 1))) {
+        if ((hubbard_) && (xi1 >= (2 * l1 + 1))) {
             break;
         }
 
         for (int xi2 = 0; xi2 < indexb.size(); xi2++) {
-            int l2  = indexb[xi2].l;
-            int lm2 = indexb[xi2].lm;
-            int o2  = indexb[xi2].order;
+            int l2                                       = indexb[xi2].l;
+            int lm2                                      = indexb[xi2].lm;
+            int o2                                       = indexb[xi2].order;
             std::array<double_complex, 3> dm_rot_spatial = {0, 0, 0};
 
             //} the hubbard treatment when spin orbit coupling is present is
@@ -573,7 +571,7 @@ inline void symmetrize(const mdarray<double_complex, 4> &ns_,
             // L.S correction within hubbard). A better option (although still
             // wrong from physics pov) would be to consider a multi orbital case.
 
-            if ((hubbard_)&&(xi2 >= (2 * l2 + 1))) {
+            if ((hubbard_) && (xi2 >= (2 * l2 + 1))) {
                 break;
             }
 
@@ -587,8 +585,7 @@ inline void symmetrize(const mdarray<double_complex, 4> &ns_,
                     for (int m4 = -l2; m4 <= l2; m4++) {
                         int lm4 = utils::lm(l2, m4);
                         int xi4 = indexb.index_by_lm_order(lm4, o2);
-                        dm_rot_spatial[j] += ns_(xi3, xi4, j, ia) *
-                            rotm(lm1, lm3) * rotm(lm2, lm4);
+                        dm_rot_spatial[j] += ns_(xi3, xi4, j, ia) * rotm(lm1, lm3) * rotm(lm2, lm4);
                     }
                 }
             }
@@ -597,9 +594,8 @@ inline void symmetrize(const mdarray<double_complex, 4> &ns_,
             if (ndm == 1) {
                 dm_(xi1, xi2, 0, ja) += dm_rot_spatial[0];
             } else {
-                double_complex spin_dm[2][2] = {
-                    {dm_rot_spatial[0], dm_rot_spatial[2]},
-                    {std::conj(dm_rot_spatial[2]), dm_rot_spatial[1]}};
+                double_complex spin_dm[2][2] = {{dm_rot_spatial[0], dm_rot_spatial[2]},
+                                                {std::conj(dm_rot_spatial[2]), dm_rot_spatial[1]}};
 
                 /* spin blocks of density matrix are: uu, dd, ud
                    the mapping from linear index (0, 1, 2) of density matrix components is:
@@ -609,7 +605,8 @@ inline void symmetrize(const mdarray<double_complex, 4> &ns_,
                 for (int k = 0; k < ndm; k++) {
                     for (int is = 0; is < 2; is++) {
                         for (int js = 0; js < 2; js++) {
-                            dm_(xi1, xi2, k, ja) += spin_rot_su2(k & 1, is) * spin_dm[is][js] * std::conj(spin_rot_su2(std::min(k, 1), js));
+                            dm_(xi1, xi2, k, ja) +=
+                                spin_rot_su2(k & 1, is) * spin_dm[is][js] * std::conj(spin_rot_su2(std::min(k, 1), js));
                         }
                     }
                 }
@@ -619,8 +616,8 @@ inline void symmetrize(const mdarray<double_complex, 4> &ns_,
 }
 
 inline void
-symmetrize(std::function<sddk::mdarray<double_complex, 3>&(int ia__)> dm__,
-           int num_mag_comp__, Crystal_symmetry const& sym__,
+symmetrize(std::function<sddk::mdarray<double_complex, 3>&(int ia__)> dm__, int num_mag_comp__,
+           Crystal_symmetry const& sym__,
            std::function<sirius::experimental::basis_functions_index const*(int)> indexb__)
 {
     /* quick exit */
@@ -646,9 +643,9 @@ symmetrize(std::function<sddk::mdarray<double_complex, 3>&(int ia__)> dm__,
 
     /* loop over symmetry operations */
     for (int isym = 0; isym < sym__.size(); isym++) {
-        int  pr   = sym__[isym].spg_op.proper;
-        auto eang = sym__[isym].spg_op.euler_angles;
-        auto rotm = sht::rotation_matrix<double>(lmax, eang, pr);
+        int pr            = sym__[isym].spg_op.proper;
+        auto eang         = sym__[isym].spg_op.euler_angles;
+        auto rotm         = sht::rotation_matrix<double>(lmax, eang, pr);
         auto spin_rot_su2 = rotation_matrix_su2(sym__[isym].spin_rotation);
 
         for (int ia = 0; ia < sym__.num_atoms(); ia++) {
@@ -669,13 +666,13 @@ symmetrize(std::function<sddk::mdarray<double_complex, 3>&(int ia__)> dm__,
             /* loop over radial functions */
             for (int idxrf1 = 0; idxrf1 < indexr.size(); idxrf1++) {
                 /* angular momentum of radial function */
-                auto am1 = indexr.am(idxrf1);
-                auto ss1 = am1.subshell_size();
+                auto am1     = indexr.am(idxrf1);
+                auto ss1     = am1.subshell_size();
                 auto offset1 = indexb.offset(idxrf1);
                 for (int idxrf2 = 0; idxrf2 < indexr.size(); idxrf2++) {
                     /* angular momentum of radial function */
-                    auto am2 = indexr.am(idxrf2);
-                    auto ss2 = am2.subshell_size();
+                    auto am2     = indexr.am(idxrf2);
+                    auto ss2     = am2.subshell_size();
                     auto offset2 = indexb.offset(idxrf2);
 
                     dm_ia.zero();
@@ -686,8 +683,8 @@ symmetrize(std::function<sddk::mdarray<double_complex, 3>&(int ia__)> dm__,
                                 for (int m1p = 0; m1p < ss1; m1p++) {
                                     for (int m2p = 0; m2p < ss2; m2p++) {
                                         dm_ia(m1, m2, j) += rotm[am1.l()](m1, m1p) *
-                                            dm__(ja)(offset1 + m1p, offset2 + m2p, j) *
-                                            rotm[am2.l()](m2, m2p);
+                                                            dm__(ja)(offset1 + m1p, offset2 + m2p, j) *
+                                                            rotm[am2.l()](m2, m2p);
                                     }
                                 }
                             }
@@ -717,8 +714,7 @@ symmetrize(std::function<sddk::mdarray<double_complex, 3>&(int ia__)> dm__,
                                     for (int s1p = 0; s1p < 2; s1p++) {
                                         for (int s2p = 0; s2p < 2; s2p++) {
                                             dmsym[ia](m1 + offset1, m2 + offset2, j) +=
-                                                spin_rot_su2(s1, s1p) * dm[s1p][s2p] *
-                                                std::conj(spin_rot_su2(s2, s2p));
+                                                spin_rot_su2(s1, s1p) * dm[s1p][s2p] * std::conj(spin_rot_su2(s2, s2p));
                                         }
                                     }
                                 }
@@ -733,7 +729,6 @@ symmetrize(std::function<sddk::mdarray<double_complex, 3>&(int ia__)> dm__,
                             }
                         }
                     }
-
                 }
             }
         }
@@ -747,11 +742,10 @@ symmetrize(std::function<sddk::mdarray<double_complex, 3>&(int ia__)> dm__,
             for (size_t i = 0; i < dm__(ia).size(); i++) {
                 dm__(ia)[i] = dmsym[ia][i] * alpha;
             }
-
         }
     }
 }
 
-} // namespace
+} // namespace sirius
 
 #endif // __SYMMETRIZE_HPP__
